﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=10
@EndOfDesignText@
' CHANGES:
' ========
' 2024-04-09:
' - initial version; wmCRUDversion (database structure version) = 1
' 2024-04-21:
' - added 'zero/empty string' provisions for filters (Note: this is for "" values, not for NULLs)
' 2024-05-04:
' - subs 'SQLtableExists' and 'SQLgetTableColumnNames' have been corrected so that they work with MySQL databases too
' 2024-05-14:
' - replaced "[...]" with "`...`" for all SQL-related strings (MySQL doesn't allow square brackets; SQLite does allow backticks)

'Prerequisites/dependencies:
'===========================
'- All files from the Assets dir whose names start with 'wmcrud' must be added to the project
'- Library dependencies (all internal libraries): ByteConverter, CSSUtils, JavaObject, jCore, jFX, jReflection, jSQL, jXUI, XUI Views
'- Additional jar requirement: sqlite-jdbc-3.39.2.0 (or another version of the sqlite-jdbc jar, see https://www.b4x.com/android/forum/threads/sqlite-jdbc-library-version-updates.133792)
'- Check the Initialize2 method's arguments for more info

Sub Class_Globals

	Private wmCRUDversion As Int = 2

	' Form et al
	Private fx As JFX
	Private MainForm As Form
	Private xui As XUI
	Private caller As Object
	Private dataValidationCallbackSub As String = ""
	Private closedCallbackSub As String = ""
	Private dbChangedCallbackSub As String = ""
	Private addedStyleSheets As Boolean = False
	Private Const dialogWidth As Double = 570dip
	Private Const dialogHeight As Double = 510dip
	Private Const borderWidth As Double = 1dip
	Private Const borderRadius As Double = 3dip
	Private Const tvCellSize As Double = 30dip ' Must be the same value as "-fx-cell-size" in file 'wmcrud_tvformat.css'
	Private isInitialized2 As Boolean = False
	Private update_add_dlg As B4XDialog

	' Context Menu
	Private contextMenu1 As ContextMenu
	Private allowNew As Boolean
	Private allowDelete As Boolean
	Private allowEdit As Boolean

	' Colours
	Private clr_backgroundInt As Int = 0xFF28282B ' Matte black
	Private clr_foregroundInt As Int = xui.Color_White
	Private clr_foregroundPaint As Paint

	' Fonts
	Private Const arialGeneralSize As Int = 13
	Private arialGeneral As Font
	Private arialBoldGeneral As Font
	Private Const arialBoldForDialogButtonsSize As Int = 11
	Private arialBoldForDialogButtons As Font
	Private Const fontAwesomeForTableViewSize As Int = 18
	Private fontAwesomeForTableView As Font
	Private Const materialIconsForTableViewSize As Int = 18
	Private materialIconsForTableView As Font

	' Arrow images for TableView headers
	Private arrowSize As Double = 40dip
	Private arrowUp As B4XBitmap
	Private arrowDown As B4XBitmap
	Private arrowRight As B4XBitmap

	' Database
	Type wmCRUDcolumnDefinition(columnHeader As String, unqualifiedColumnName As String, qualifiedColumnName As String, alignment As String, _
								allowEmpty As Int, allowZero As Int, allowEdit As Int, _
								dataType As Int, comboBoxItemsValues As List, comboBoxItemsDisplay As List)
	Private columnDefinitions As List
	Private columnDefinitionsOrderedLikeTV As List
	Private mapColDefsCBlist As Map ' Key = wmCRUDcolumnDefinition.qualifiedColumnName, value=a list with 2 elements: wmCRUDcolumnDefinition.comboBoxItemsDisplay, wmCRUDcolumnDefinition.comboBoxItemsValues
	Type wmCRUDlabelTagType(ID As Int) ' Using a Type instead of a simple variable in case there would be a need for more members
	Private dataDir As String
	Private uniqueIDcolumnName As String
	Private sqlMeta As SQL
	Private sql1 As SQL
	Private tablename As String = "--"
	Private tableDisplayName As String
	Private dbWasChanged As Boolean = False
	Private dbSortColumn As String = ""
	Private dbSortOrder As String = ""
	Private dbWhereFilter As String = ""
	Private dbWhereFilterValuesList As List
	Private dbWhereFilterViewList As List
	Private cannotDeleteIDvalues As List
	Private cannotEditIDvalues As List

	Private mapDataTypes As Map

	Private Const WMCRUD_DATATYPE_TEXT As Int = 0
	Private Const WMCRUD_DATATYPE_MULTILINE_TEXT As Int = 1
	Private Const WMCRUD_DATATYPE_INT As Int = 2
	Private Const WMCRUD_DATATYPE_INT_POS As Int = 3
	Private Const WMCRUD_DATATYPE_INT_NEG As Int = 4
	Private Const WMCRUD_DATATYPE_DEC As Int = 5
	Private Const WMCRUD_DATATYPE_DEC_POS As Int = 6
	Private Const WMCRUD_DATATYPE_DEC_NEG As Int = 7
	Private Const WMCRUD_DATATYPE_COMBOBOX As Int = 8

	Private Const WMCRUD_DATATYPE_DATE_START As Int = 100
	Private Const WMCRUD_DATATYPE_DATE_YYYYMMDD As Int = 101
	Private Const WMCRUD_DATATYPE_DATE_LONG As Int = 102
	Private Const WMCRUD_DATATYPE_DATE_YYYYMMDD_HYPHENATED As Int = 103
	Private Const WMCRUD_DATATYPE_DATE_END As Int = 199

	Private Const WMCRUD_DATATYPE_WITH_FONTAWESOME_START As Int = 1000
	Private Const WMCRUD_DATATYPE_TEXT_WITH_FONTAWESOME As Int = 1001
	Private Const WMCRUD_DATATYPE_INT_WITH_FONTAWESOME As Int = 1002
	Private Const WMCRUD_DATATYPE_INT_POS_WITH_FONTAWESOME As Int = 1003
	Private Const WMCRUD_DATATYPE_INT_NEG_WITH_FONTAWESOME As Int = 1004
	Private Const WMCRUD_DATATYPE_DEC_WITH_FONTAWESOME As Int = 1005
	Private Const WMCRUD_DATATYPE_DEC_POS_WITH_FONTAWESOME As Int = 1006
	Private Const WMCRUD_DATATYPE_DEC_NEG_WITH_FONTAWESOME As Int = 1007
	Private Const WMCRUD_DATATYPE_WITH_FONTAWESOME_END As Int = 1999

	Private Const WMCRUD_DATATYPE_WITH_MATERIALICONS_START As Int = 2000
	Private Const WMCRUD_DATATYPE_TEXT_WITH_MATERIALICONS As Int = 2001
	Private Const WMCRUD_DATATYPE_INT_WITH_MATERIALICONS As Int = 2002
	Private Const WMCRUD_DATATYPE_INT_POS_WITH_MATERIALICONS As Int = 2003
	Private Const WMCRUD_DATATYPE_INT_NEG_WITH_MATERIALICONS As Int = 2004
	Private Const WMCRUD_DATATYPE_DEC_WITH_MATERIALICONS As Int = 2005
	Private Const WMCRUD_DATATYPE_DEC_POS_WITH_MATERIALICONS As Int = 2006
	Private Const WMCRUD_DATATYPE_DEC_NEG_WITH_MATERIALICONS As Int = 2007
	Private Const WMCRUD_DATATYPE_WITH_MATERIALICONS_END As Int = 2999

	' This is used to reduce the number of members shown by the IDE for a wmCRUD variable.
	' This way, instead of seeing all the 'WMCRUD_DATATYPE_...' members (if they were Public),
	' only member 'DataTypes' is shown, which then contains the below members.
	Type wmCRUDdataTypes(TEXT As Int, MULTILINE_TEXT As Int, _
						INTEGER As Int, INTEGER_POSITIVE As Int, INTEGER_NEGATIVE As Int, _
						DECIMAL As Int, DECIMAL_POSITIVE As Int, DECIMAL_NEGATIVE As Int, _
						COMBOBOX As Int, _
						DATE_YYYYMMDD As Int, DATE_LONG As Int, DATE_YYYYMMDD_HYPHENATED As Int, _
						TEXT_WITH_FONTAWESOME As Int, _
						INTEGER_WITH_FONTAWESOME As Int, INTEGER_POSITIVE_WITH_FONTAWESOME As Int, INTEGER_NEGATIVE_WITH_FONTAWESOME As Int, _
						DECIMAL_WITH_FONTAWESOME As Int, DECIMAL_POSITIVE_WITH_FONTAWESOME As Int, DECIMAL_NEGATIVE_WITH_FONTAWESOME As Int, _
						TEXT_WITH_MATERIALICONS As Int, _
						INTEGER_WITH_MATERIALICONS As Int, INTEGER_POSITIVE_WITH_MATERIALICONS As Int, INTEGER_NEGATIVE_WITH_MATERIALICONS As Int, _
						DECIMAL_WITH_MATERIALICONS As Int, DECIMAL_POSITIVE_WITH_MATERIALICONS As Int, DECIMAL_NEGATIVE_WITH_MATERIALICONS As Int)
	Public DataTypes As wmCRUDdataTypes

	' Number separators
	Private thousandsSeparator As String
	Private decimalSeparator As String

	' Filters
	Private myShowFilters As Boolean = True
	Private filterFieldHeight As Double = 30dip
	Private lblFilters As Label
	Private TV1SetColumnsInProgress As Boolean = False
	Private TV1originalLeft As Double

	' Miscellaneous
	Private fontAwesomeMap As Map ' FontAwesome icons for the TableView, based on https://www.b4x.com/android/forum/threads/fontawesome-at-runtime.85628/
	Private materialIconsMap As Map ' MaterialIcons for the TableView
	Private showingData As Boolean = False
	Private showingDataAfterSettingFilter As Boolean = False

	' editPane; at the moment, this isn't used and hasn't been tested
	Private editPane As B4XView
	Private mapEditPaneViews As Map ' Key: the unqualified column name; value: the view

	' Helper variables for wmCRUDTableView1
	Private resizedColumnsList As List
	Private resizedColumnsTimer As Timer
	Private resizingColumns As Boolean = False

	' Designer Views
	Private wmCRUDPaneStatus As Pane
	Private wmCRUDLabelBack As Label
	Private wmCRUDLabelStatus As Label
	Private wmCRUDLabelRefresh As Label
	Private wmCRUDPaneContent As Pane
	Private wmCRUDLabelWorking As Label
	Private wmCRUDPaneFilter As Pane
	Private wmCRUDTableView1 As TableView
	Private wmCRUDPaneAddUpdate As Pane
	Type wmCRUDdataValidation(instance As wmCRUD, dlg As B4XDialog, btnSave As B4XView, btnSaveNew As B4XView, btnCancel As B4XView, errorLabel As Label, columnViews As List, columnLabels As List)
	Private updateOrAddDataValidation As wmCRUDdataValidation
	Private fillingDialog As Boolean = False

	' Localisation; when putting new values, the keys are case-sensitive!
	Public LocalisedStrings As Map = CreateMap("(none)": "", _
												"A value is required here": "", _
												"A non-zero value is required here": "", _
												"Back": "", _
												"Cancel": "", _
												"Click to reset all filters": "", _
												"Delete": "", _
												"Duplicate": "", _
												"Edit": "", _
												"Enter (part of) a value and press ENTER to apply this filter": "", _
												"Error": "", _
												"Filters": "", _
												"From": "", _
												"New": "", _
												"No": "", _
												"No data found": "", _
												"OK to delete?": "", _
												"Please confirm": "", _
												"Refresh": "", _
												"Rows showing": "", _
												"Save": "", _
												"Save new": "", _
												"This must be a decimal number": "", _
												"This must be a negative decimal number": "", _
												"This must be a negative integer number": "", _
												"This must be a positive decimal number": "", _
												"This must be a positive integer number": "", _
												"This must be an integer number": "", _
												"Use two double quotes ("") to filter on empty text values, or zero for empty numeric values": "", _
												"To": "", _
												"Yes": "", _
												"Working...": "")

End Sub

#Region Initalisations and settings
' This method only initialises the Class so that MakeColumnDefinition can be used.
' After having used MakeColumnDefinition, a call to Initialize2 must follow.
Public Sub Initialize

	' Icons for the TableView, based on https://www.b4x.com/android/forum/threads/fontawesome-at-runtime.85628/
	fontAwesomeMap.Initialize
	materialIconsMap.Initialize

	' Data types
	mapDataTypes.Initialize

	mapDataTypes.Put(WMCRUD_DATATYPE_TEXT, "t")
	mapDataTypes.Put(WMCRUD_DATATYPE_MULTILINE_TEXT, "tm")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_POS, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_NEG, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_POS, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_NEG, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_COMBOBOX, "cb")
	mapDataTypes.Put(WMCRUD_DATATYPE_DATE_YYYYMMDD, "dtymd")
	mapDataTypes.Put(WMCRUD_DATATYPE_DATE_LONG, "dtlong")
	mapDataTypes.Put(WMCRUD_DATATYPE_DATE_YYYYMMDD_HYPHENATED, "dtymdh")

	mapDataTypes.Put(WMCRUD_DATATYPE_TEXT_WITH_FONTAWESOME, "t")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_WITH_FONTAWESOME, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_POS_WITH_FONTAWESOME, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_NEG_WITH_FONTAWESOME, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_WITH_FONTAWESOME, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_POS_WITH_FONTAWESOME, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_NEG_WITH_FONTAWESOME, "d")

	mapDataTypes.Put(WMCRUD_DATATYPE_TEXT_WITH_MATERIALICONS, "t")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_WITH_MATERIALICONS, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_POS_WITH_MATERIALICONS, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_INT_NEG_WITH_MATERIALICONS, "i")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_WITH_MATERIALICONS, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_POS_WITH_MATERIALICONS, "d")
	mapDataTypes.Put(WMCRUD_DATATYPE_DEC_NEG_WITH_MATERIALICONS, "d")

	DataTypes.Initialize

	DataTypes.TEXT = WMCRUD_DATATYPE_TEXT
	DataTypes.MULTILINE_TEXT = WMCRUD_DATATYPE_MULTILINE_TEXT
	DataTypes.INTEGER = WMCRUD_DATATYPE_INT
	DataTypes.INTEGER_POSITIVE = WMCRUD_DATATYPE_INT_POS
	DataTypes.INTEGER_NEGATIVE = WMCRUD_DATATYPE_INT_NEG
	DataTypes.DECIMAL = WMCRUD_DATATYPE_DEC
	DataTypes.DECIMAL_POSITIVE = WMCRUD_DATATYPE_DEC_POS
	DataTypes.DECIMAL_NEGATIVE = WMCRUD_DATATYPE_DEC_NEG
	DataTypes.COMBOBOX = WMCRUD_DATATYPE_COMBOBOX
	DataTypes.DATE_YYYYMMDD = WMCRUD_DATATYPE_DATE_YYYYMMDD
	DataTypes.DATE_LONG = WMCRUD_DATATYPE_DATE_LONG
	DataTypes.DATE_YYYYMMDD_HYPHENATED = WMCRUD_DATATYPE_DATE_YYYYMMDD_HYPHENATED

	DataTypes.TEXT_WITH_FONTAWESOME = WMCRUD_DATATYPE_TEXT_WITH_FONTAWESOME
	DataTypes.INTEGER_WITH_FONTAWESOME = WMCRUD_DATATYPE_INT_WITH_FONTAWESOME
	DataTypes.INTEGER_POSITIVE_WITH_FONTAWESOME = WMCRUD_DATATYPE_INT_POS_WITH_FONTAWESOME
	DataTypes.INTEGER_NEGATIVE_WITH_FONTAWESOME = WMCRUD_DATATYPE_INT_NEG_WITH_FONTAWESOME
	DataTypes.DECIMAL_WITH_FONTAWESOME = WMCRUD_DATATYPE_DEC_WITH_FONTAWESOME
	DataTypes.DECIMAL_POSITIVE_WITH_FONTAWESOME = WMCRUD_DATATYPE_DEC_POS_WITH_FONTAWESOME
	DataTypes.DECIMAL_NEGATIVE_WITH_FONTAWESOME = WMCRUD_DATATYPE_DEC_NEG_WITH_FONTAWESOME

	DataTypes.TEXT_WITH_MATERIALICONS = WMCRUD_DATATYPE_TEXT_WITH_MATERIALICONS
	DataTypes.INTEGER_WITH_MATERIALICONS = WMCRUD_DATATYPE_INT_WITH_MATERIALICONS
	DataTypes.INTEGER_POSITIVE_WITH_MATERIALICONS = WMCRUD_DATATYPE_INT_POS_WITH_MATERIALICONS
	DataTypes.INTEGER_NEGATIVE_WITH_MATERIALICONS = WMCRUD_DATATYPE_INT_NEG_WITH_MATERIALICONS
	DataTypes.DECIMAL_WITH_MATERIALICONS = WMCRUD_DATATYPE_DEC_WITH_MATERIALICONS
	DataTypes.DECIMAL_POSITIVE_WITH_MATERIALICONS = WMCRUD_DATATYPE_DEC_POS_WITH_MATERIALICONS
	DataTypes.DECIMAL_NEGATIVE_WITH_MATERIALICONS = WMCRUD_DATATYPE_DEC_NEG_WITH_MATERIALICONS

End Sub

' Use this method to generate wmCRUDcolumnDefinition entries for list columnDefinitions that will be passed to Initialize2.
' It can be called after Initialize and before Initialize2 is called.
' One wmCRUDcolumnDefinition variable should be created for each column in the table; missing columns are assigned default values.
' Arguments:
' - columnHeader: column header text
' - columnName: column name in the database
' - alignment: LEFT / RIGHT / CENTER
' - allowEditing: True or False to indicate whether editing is not allowed for that column
' - allowEmpty: True or False to indicate whether an empty value is allowed for that column
' - allowZero: True or False to indicate whether a zero value is allowed for that column
' - dataType: one of the values in DataTypes
' - comboBoxItemsDisplay: a List with the display values for the corresponding filter field's Combobox, or Null
' - comboBoxItemsValues: a List with the database table values corresponding to the comboBoxItemsDisplay values
' Hint: for boolean data, use dataType WMCRUD_DATATYPE_COMBOBOX, and use vlues like Yes/No, True/False, 1/0
Public Sub MakeColumnDefinition(columnHeader As String, columnName As String, alignment As String, _
								allowEmpty As Boolean, allowZero As Boolean, allowEditing As Boolean, _
								dataType As Int, comboBoxItemsValues As List, comboBoxItemsDisplay As List) As wmCRUDcolumnDefinition

	Dim t1 As wmCRUDcolumnDefinition
	t1.Initialize
	t1.alignment = alignment
	t1.allowEdit = IIf(allowEditing, 1, 0)
	t1.allowEmpty = IIf(allowEmpty, 1, 0)
	t1.allowZero = IIf(allowZero, 1, 0)
	If comboBoxItemsDisplay.IsInitialized And (comboBoxItemsDisplay <> Null) Then t1.comboBoxItemsDisplay = comboBoxItemsDisplay
	If comboBoxItemsValues.IsInitialized And (comboBoxItemsValues <> Null) Then t1.comboBoxItemsValues = comboBoxItemsValues
	t1.columnHeader = columnHeader
	t1.unqualifiedColumnName = columnName
	t1.qualifiedColumnName = "" ' This will be filled in PrepareColumnDefinitionsAndMetaDb
	t1.dataType = dataType
	Return t1

End Sub

'Must be called after Initialize and (optionally but recommended) MakeColumnDefinition calls. Check the log/stdout if it returns False.
'Arguments:
'- caller1: the calling Module
'- form1: the calling Form
'- parentPane1: the Pane into which the Class will load its layout
'- dataDir1: the directory where SQLite database 'wmCRUD_meta.sqlite' will be stored
'- sqlDb As SQL: the (already opened) SQLite (or other SQL - not tested) database
'- tableName1: the name of the database table (also see 'uniqueIDcolumnName1')
'- tableDisplayName1: the name to show in the class's status label at the top
'- uniqueIDcolumnName1: the name of the INTEGER AUTOINCREMENT column in the database table that contains the unique ID for each row
'- cannotDeleteIDvalues1: a List with unique ID values that are not allowed to be deleted
'- cannotEditIDvalues1: a List with unique ID values that are not allowed to be edited
'- columnDefinitions1: a List of wmCRUDcolumnDefinition variables; create them with calls to method MakeColumnDefinition; this can be done before Initialize2 is called
'- allowNew1: if True, the ContextMenu (when right-clicking a row) will contain the 'Add' and 'Duplicate' options
'- allowEdit1: if True, the ContextMenu (when right-clicking a row) will contain the 'Edit' option, and double-clicking a row in the TableView will allow editing it as well
'- allowDelete1: if True, the ContextMenu (when right-clicking a row) will contain the 'Delete' option
'- dataValidationCallbackSub1: if not "", the method in caller1 to be used for data validations while editing a row; it must have an argument like: 'dv As wmCRUDdataValidation'
'- dbChangedCallbackSub1: if not "", the method in caller1 to be called when database data have been changed; it must have two arguments like: 'instance As wmCRUD, changeType As String'; changeType is 'c', 'u', or 'd'
'- closedCallbackSub1: if not "", the method in caller1 to be called when the 'Back' arrow is clicked; it must have an argument like: 'instance As wmCRUD'
Public Sub Initialize2(caller1 As Object, form1 As Form, parentPane1 As Pane, _
						dataDir1 As String, sqlDb As SQL, _
						tableName1 As String, tableDisplayName1 As String, uniqueIDcolumnName1 As String, _
						cannotDeleteIDvalues1 As List, cannotEditIDvalues1 As List, _
						columnDefinitions1 As List, _
						allowNew1 As Boolean, allowEdit1 As Boolean, allowDelete1 As Boolean, _
						dataValidationCallbackSub1 As String, dbChangedCallbackSub1 As String, closedCallbackSub1 As String) As Boolean

	Dim missingColumns As Boolean = False

	' Table name
	tablename = tableName1
	tableDisplayName = tableDisplayName1

	' Caller, Form, ContextMenu, and wmUtils
	caller = caller1
	MainForm = form1

	If FileExistsInDirAssets("wmcrud_tvformat.css") = False Then
		Log("wmCRUD.Initialize2 (" & tablename & ") - file 'wmcrud_tvformat.css' is missing from the Assets directory")
		DoClose(True)
		Return False
	Else If MainForm.Stylesheets.IndexOf(File.GetUri(File.DirAssets, "wmcrud_tvformat.css")) < 0 Then
		addedStyleSheets = True
		MainForm.Stylesheets.Add(File.GetUri(File.DirAssets, "wmcrud_tvformat.css")) ' To make the labels fill the entire cells without padding around them, and other formatting
		MainForm.Stylesheets.Add(File.GetUri(File.DirAssets, "wmcrud_comboboxdarktheme.css"))
	End If

	If FileExistsInDirAssets("wmcrud.bjl") = False Then
		Log("wmCRUD.Initialize2 (" & tablename & ") - file 'wmcrud.bjl' is missing from the Assets directory")
		DoClose(True)
		Return False
	End If

	parentPane1.LoadLayout("wmcrud.bjl")
	TV1originalLeft = wmCRUDTableView1.Left
	wmCRUDLabelWorking.Text = Localise("Working...")
	contextMenu1.Initialize("ContextMenu1")
	SetTableViewPlaceHolderText(wmCRUDTableView1, Localise("No data found"))

	' Callback subs
	dataValidationCallbackSub = dataValidationCallbackSub1
	If dataValidationCallbackSub <> "" Then
		If SubExists(caller, dataValidationCallbackSub) = False Then
			Log("wmCRUD.Initialize2 (" & tablename & ") - dataValidationCallbackSub1 doesn't exist in caller1: " & dataValidationCallbackSub1)
			closedCallbackSub = ""
			DoClose(True)
			Return False
		End If
	End If

	closedCallbackSub = closedCallbackSub1
	If closedCallbackSub <> "" Then
		If SubExists(caller, closedCallbackSub) = False Then
			Log("wmCRUD.Initialize2 (" & tablename & ") - closedCallbackSub1 doesn't exist in caller1: " & closedCallbackSub1)
			closedCallbackSub = ""
			DoClose(True)
			Return False
		End If
	End If

	dbChangedCallbackSub = dbChangedCallbackSub1
	If dbChangedCallbackSub <> "" Then
		If SubExists(caller, dbChangedCallbackSub) = False Then
			Log("wmCRUD.Initialize2 (" & tablename & ") - dbChangedCallbackSub1 doesn't exist in caller1: " & dbChangedCallbackSub1)
			closedCallbackSub = ""
			DoClose(True)
			Return False
		End If
	End If

	' Code from https://www.b4x.com/android/forum/threads/sorting-tableview-custom-columns.65292/
	' and https://www.b4x.com/android/forum/threads/tableview-resize-column-event.67650/#post-428670
	Dim r As Reflector
	r.Target = wmCRUDTableView1
	r.AddEventFilter("wmCRUDTableView1MouseAny", "javafx.scene.input.MouseEvent.ANY")
	r.AddEventFilter("wmCRUDTableView1MouseClicked", "javafx.scene.input.MouseEvent.MOUSE_CLICKED")

	' Code based on https://www.b4x.com/android/forum/threads/tableview-multiple-selection-via-program.100930/#post-634304
	Dim wmCRUDTableView1SelectionModel As JavaObject
	Dim jo As JavaObject = wmCRUDTableView1
	Dim joSelMode As JavaObject
	joSelMode.InitializeStatic("javafx.scene.control.SelectionMode")
	'get Selection Model:
	wmCRUDTableView1SelectionModel = jo.RunMethodJO("getSelectionModel", Null)
	'set the selectionmode of the tableview by using the enum
	wmCRUDTableView1SelectionModel.RunMethod("setSelectionMode", Array(joSelMode.GetField("MULTIPLE")))

	' Colours
	clr_foregroundPaint = fx.Colors.From32Bit(clr_foregroundInt)

	' Fonts
	fontAwesomeForTableView = fx.CreateFontAwesome(fontAwesomeForTableViewSize)
	materialIconsForTableView = fx.CreateMaterialIcons(materialIconsForTableViewSize)
	arialGeneral = fx.CreateFont("Arial", arialGeneralSize, False, False)
	arialBoldGeneral = fx.CreateFont("Arial", arialGeneralSize, True, False)
	arialBoldForDialogButtons = fx.CreateFont("Arial", arialBoldForDialogButtonsSize, True, False)

	' Number separators
	decimalSeparator = NumberFormat2(1.1, 1, 1, 1, False).SubString2(1, 2)
	thousandsSeparator = NumberFormat2(1000, 1, 0, 0, True).SubString2(1, 2)

	' Arrow images for TableView headers
	arrowUp = xui.LoadBitmapResize(File.DirAssets, "wmcrud_arrow_up_black.png", arrowSize, arrowSize, True)
	arrowDown = xui.LoadBitmapResize(File.DirAssets, "wmcrud_arrow_down_black.png", arrowSize, arrowSize, True)
	arrowRight = xui.LoadBitmapResize(File.DirAssets, "wmcrud_arrow_right_black.png", arrowSize, arrowSize, True)

	' Allowed operations
	allowNew = allowNew1
	allowDelete = allowDelete1
	allowEdit = allowEdit1

	' Database
	If (File.Exists(dataDir1, "") = False) Or (File.IsDirectory(dataDir1, "") = False) Then
		Log("wmCRUD.Initialize2 (" & tablename & ") - dataDir1 doesn't exist or isn't a directory: " & dataDir1)
		DoClose(True)
		Return False
	Else
		dataDir = dataDir1
	End If
	sql1 = sqlDb
	If SQLtableExists(sql1, tablename) = False Then
		Log("wmCRUD.Initialize2 (" & tablename & ") - tableName1 doesn't exist: " & tablename)
		DoClose(True)
		Return False
	End If
	uniqueIDcolumnName = uniqueIDcolumnName1
	If SQLcolumnExists(sql1, tablename, uniqueIDcolumnName) = False Then
		Log("wmCRUD.Initialize2 (" & tablename & ") - uniqueIDcolumnName1 '" & uniqueIDcolumnName & "' doesn't exist in table '" & tablename & "'")
		DoClose(True)
		Return False
	End If

	missingColumns = False
	For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitions1
		If SQLcolumnExists(sql1, tablename, oneColDef.unqualifiedColumnName) = False Then
			Log("wmCRUD.Initialize2 (" & tablename & ") - column '" & oneColDef.unqualifiedColumnName & "' doesn't exist in table '" & tablename & "'")
			missingColumns = True
		End If
	Next
	If missingColumns Then
		DoClose(True)
		Return False
	End If

	' Ensure that all header values in columnDefinitions1 are unique.
	' Duplicate headers cause issues when looking up the db column name via the header.
	Dim headersMap As Map
	headersMap.Initialize
	For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitions1
		If headersMap.ContainsKey(oneColDef.columnHeader.ToLowerCase) Then
			Log("wmCRUD.Initialize2 (" & tablename & ") - duplicate column header found in columnDefinitions1: " & oneColDef.columnHeader)
			DoClose(True)
			Return False
		Else
			headersMap.put(oneColDef.columnHeader.ToLowerCase, "")
		End If
	Next

	' editPane - I decided not to use this but left the code in the class;
	' it has not been tested and is definitely incomplete now.
	mapEditPaneViews.Initialize
'	If editPane1.IsInitialized And (editPane1 <> Null) Then
'		editPane = editPane1
'		For Each n As Node In editPane.GetAllViewsRecursive
'			If n.Id <> "" Then mapEditPaneViews.Put(n.Id.ToLowerCase, n)
'		Next
'		missingColumns = False
'		For Each s As String In SQLgetTableColumnNames(sql1, tablename, False)
'			If mapEditPaneViews.ContainsKey(s.ToLowerCase) = False Then
'				Log("wmCRUD.Initialize2 (" & tablename & ") - database column not found as 'view.id' in editPane1: " & s)
'				missingColumns = True
'			End If
'		Next
'		If missingColumns Then
'			DoClose(True)
'			Return False
'		End If
'	End If

	' Update/Add dialog
	update_add_dlg.Initialize(MainForm.RootPane)
	update_add_dlg.BackgroundColor = clr_backgroundInt
	update_add_dlg.BlurBackground = True
	update_add_dlg.BorderColor = clr_foregroundInt
	update_add_dlg.ButtonsColor = xui.Color_LightGray
	update_add_dlg.ButtonsFont = arialBoldForDialogButtons
	update_add_dlg.ButtonsTextColor = xui.Color_Black
	update_add_dlg.TitleBarColor = clr_backgroundInt
	update_add_dlg.TitleBarFont = arialBoldGeneral
	update_add_dlg.TitleBarTextColor = clr_foregroundInt

	' Lists and column definitions
	If cannotDeleteIDvalues1.IsInitialized Then
		cannotDeleteIDvalues = CopyList(cannotDeleteIDvalues1, False)
	Else
		Dim emptyList As List
		emptyList.Initialize
		cannotDeleteIDvalues = emptyList
	End If
	If cannotEditIDvalues1.IsInitialized Then
		cannotEditIDvalues = CopyList(cannotEditIDvalues1, False)
	Else
		Dim emptyList As List
		emptyList.Initialize
		cannotEditIDvalues = emptyList
	End If
	columnDefinitions = CopyList(columnDefinitions1, False)
	mapColDefsCBlist.Initialize
	PrepareColumnDefinitionsAndMetaDb

	' Ready
	isInitialized2 = True
	Return True

End Sub

' Resets all saved layout data (column order and widths, sort order, and filter values).
' If 'andShow' is True, method 'Show' is called too.
Public Sub ResetLayout(andShow As Boolean)

	If isInitialized2 = False Then
		Log("wmCRUD.ResetLayout - Initialize2 has not been called")
		Return
	End If

	sqlMeta.ExecNonQuery2("DELETE FROM `TableViewColumns` WHERE `TableName`=?", Array As String(tablename))
	PrepareColumnDefinitionsAndMetaDb
	If andShow Then Show

End Sub

' Sets foreground and background colours. Call this method before calling Show.
Public Sub SetColors(background As Int, foreground As Int)

	If isInitialized2 = False Then
		Log("wmCRUD.SetColors - Initialize2 has not been called")
		Return
	End If

	clr_backgroundInt = background
	clr_foregroundInt = foreground
	clr_foregroundPaint = fx.Colors.From32Bit(foreground)

End Sub

Private Sub wmCRUDLabelBack_MouseClicked (EventData As MouseEvent)

	If isInitialized2 Then Close

End Sub

Private Sub wmCRUDLabelRefresh_MouseClicked (EventData As MouseEvent)

	Show

End Sub

' Saves the layout (column order and widths) in the meta database and hides the views.
' To continue the class after calling Close, Initialize2 must be called again.
Public Sub Close

	If isInitialized2 = False Then
		Log("wmCRUD.Close - Initialize2 has not been called")
		Return
	End If

	DoClose(False)

End Sub

Private Sub DoClose(calledByInitialize2 As Boolean)

	If calledByInitialize2 = False Then
		TV1SaveColumnWidthsAndOrder(True)
		SetWaitCursor(False)
	End If

	If wmCRUDLabelWorking.IsInitialized Then wmCRUDLabelWorking.Visible = False
	If wmCRUDPaneContent.IsInitialized Then wmCRUDPaneContent.Visible = False
	If wmCRUDPaneStatus.IsInitialized Then wmCRUDPaneStatus.Visible = False

	If addedStyleSheets And MainForm.IsInitialized Then
		Dim i As Int = MainForm.Stylesheets.IndexOf(File.GetUri(File.DirAssets, "wmcrud_tvformat.css"))
		MainForm.Stylesheets.RemoveAt(i)
		i = MainForm.Stylesheets.IndexOf(File.GetUri(File.DirAssets, "wmcrud_comboboxdarktheme.css"))
		MainForm.Stylesheets.RemoveAt(i)
	End If

	isInitialized2 = False

	If closedCallbackSub <> "" Then CallSub2(caller, closedCallbackSub, Me)

End Sub

' Update comboBoxItemsValues and comboBoxItemsDisplay for the given column.
' This can be useful if the caller decides to change the allowed values for a column that is represented by a ComboBox.
' Arguments:
' - columnName: column name in the database table to identify the column whose lists to update
' - comboBoxItemsDisplay: a List with the display values for the corresponding filter field's Combobox
' - comboBoxItemsValues: a List with the database table values corresponding to the comboBoxItemsDisplay values
Public Sub UpdateComboBoxValues(columnName As String, comboBoxItemsValues As List, comboBoxItemsDisplay As List)

	If isInitialized2 = False Then
		Log("wmCRUD.UpdateComboBoxValues - Initialize2 has not been called")
		Return
	End If

	Dim found As Boolean = False
	Dim colName As String = ("`" & tablename & "`.`" & columnName & "`").As(String).ToLowerCase

	For Each s As String In mapColDefsCBlist.Keys
		If s.EqualsIgnoreCase(colName) Then
			found = True
			' Prepare the data for the map
			Dim lst As List
			lst.Initialize
			lst.Add(CopyList(comboBoxItemsDisplay, True))
			lst.Add(CopyList(comboBoxItemsValues, True))
			' Update the map and exit the loop
			mapColDefsCBlist.Put(s, lst)
			Exit
		End If
	Next

	If found = False Then Log("wmCRUD.UpdateComboBoxValues (" & tablename & ") - columnName not found: " & columnName)

End Sub

' Gets or sets whether the filters pane should be produced (default=True). After changing it, call method Show to apply it.
Public Sub getShowFilters As Boolean
	Return myShowFilters
End Sub
' Gets or sets whether the filters pane should be produced (default=True). After changing it, call method Show to apply it.
Public Sub setShowFilters(newval As Boolean)
	myShowFilters = newval
End Sub

' Returns a Boolean indicating whether or not the database has been changed.
Public Sub getDbChanged As Boolean
	Return dbWasChanged
End Sub

' Returns a string containing the table name that was passed to Initialize2.
Public Sub getSelectedTable As String
	Return tablename
End Sub

Private Sub PrepareColumnDefinitionsAndMetaDb

	Dim rs As ResultSet
	Dim s As String
	Dim s2 As String
	Dim colNum As Int
	Dim dbVersion As Int
	Dim found As Boolean

	' Setup wmCRUD_meta.sqlite and its table TableViewColumns
#If DEBUG
	Log("wmCRUD.PrepareColumnDefinitionsAndMetaDb (" & tablename & ") - sqlMeta file is: " & File.Combine(dataDir, "wmCRUD_meta.sqlite"))
#End If

	For Each cd As wmCRUDcolumnDefinition In columnDefinitions
		cd.qualifiedColumnName = "`" & tablename & "`.`" & cd.unqualifiedColumnName & "`"
	Next

	sqlMeta.InitializeSQLite(dataDir, "wmCRUD_meta.sqlite", True)
	If SQLtableExists(sqlMeta, "TableViewColumns") = False Then
		sqlMeta.ExecNonQuery("CREATE TABLE `Settings` (Version INTEGER);")
		sqlMeta.ExecNonQuery2("INSERT INTO `Settings` (`Version`) VALUES (?)", Array(wmCRUDversion))
		sqlMeta.ExecNonQuery($"CREATE TABLE `TableViewColumns` (TableName STRING,
																QualifiedColumnName STRING, UnqualifiedColumnName STRING, HeaderText STRING,
																Position INTEGER, Align STRING, SortOrder STRING, Width REAL,
																DataType INTEGER,
																Filtering INTEGER, FilterValueFrom STRING, FilterValueTo STRING,
																AllowEdit INTEGER, AllowEmpty INTEGER, AllowZero INTEGER,
																PRIMARY KEY (TableName, QualifiedColumnName));"$)
		sqlMeta.ExecNonQuery("CREATE INDEX TVCidx1 ON `TableViewColumns`(`TableName`, `QualifiedColumnName`);")
		sqlMeta.ExecNonQuery("CREATE INDEX TVCidx2 ON `TableViewColumns`(`Position`);")
		sqlMeta.ExecNonQuery("CREATE INDEX TVCidx3 ON `TableViewColumns`(`TableName`, `HeaderText`);")
	Else
		' Update the existing meta database (structure and/or data), if necessary
		dbVersion = sqlMeta.ExecQuerySingleResult("SELECT `Version` FROM `Settings` LIMIT 1").As(Int)
		If (dbVersion = 1) And (wmCRUDversion > 1) Then
			' Replaced "[...]" with "`...`" for all SQL-related strings (MySQL doesn't allow square brackets; SQLite does allow backticks)
			found = True
			Do While found
				found = False
				rs = sqlMeta.ExecQuery2("SELECT `QualifiedColumnName` FROM `TableViewColumns` WHERE `QualifiedColumnName` LIKE ? LIMIT 1", Array As String("[%"))
				If rs.NextRow Then
					found = True
					s = rs.GetString("QualifiedColumnName")
					s2 = s.Replace("[", "`").Replace("]", "`")
					sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `QualifiedColumnName`=? WHERE `QualifiedColumnName`=?", Array As String(s2, s))
				End If
				rs.Close
			Loop
			' Ready, set the meta db's version number
			sqlMeta.ExecNonQuery2("UPDATE `Settings` SET `Version`=?", Array As String(wmCRUDversion))
		End If
	End If

	' Remove columns from `TableViewColumns` that aren't present in columnDefinitions.
	' columnDefinitions is a List of wmCRUDcolumnDefinition variables passed to Initialize2 and created with method MakeColumnDefinition.
	rs = sqlMeta.ExecQuery2("SELECT `QualifiedColumnName` FROM `TableViewColumns` WHERE `TableName`=?", Array As String(tablename))
	Do While rs.NextRow
		Dim foundInColDesc As Boolean = False
		For i = (columnDefinitions.Size - 1) To 0 Step -1
			Dim cd As wmCRUDcolumnDefinition = columnDefinitions.Get(i).As(wmCRUDcolumnDefinition)
			If rs.GetString("QualifiedColumnName") = cd.qualifiedColumnName Then
				foundInColDesc = True
				Exit
			End If
		Next
		If foundInColDesc = False Then
			sqlMeta.ExecNonQuery2("DELETE FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, rs.GetString("QualifiedColumnName")))
		End If
	Loop
	rs.Close

	' Ensure that all columns from columnDefinitions are present in `TableViewColumns`
	' with the correct values. After that, we can work with `TableViewColumns`.
	s = $"INSERT OR IGNORE INTO `TableViewColumns`
			(`TableName`, `QualifiedColumnName`, `UnqualifiedColumnName`, `HeaderText`,
			`Position`, `Align`, `Sortorder`, `Width`,
			`DataType`,
			`Filtering`, `FilterValueFrom`, `FilterValueTo`,
			`AllowEdit`, `AllowEmpty`, `AllowZero`)
			VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"$
	s2 = $"UPDATE `TableViewColumns` SET
			`HeaderText`=?, `Align`=?,
			`DataType`=?,
			`AllowEdit`=?, `AllowEmpty`=?, `AllowZero`=? 
			WHERE `TableName`=? AND `QualifiedColumnName`=?"$
	For i = 0 To (columnDefinitions.Size - 1)
		Dim oneColDescr As wmCRUDcolumnDefinition = columnDefinitions.Get(i)
		' Save the ComboBox items data in mapColDefsCBlist, as they're volatile and are not stored in the db
		Dim cbLst As List
		cbLst.Initialize
		cbLst.Add(CopyList(oneColDescr.comboBoxItemsDisplay, True))
		cbLst.Add(CopyList(oneColDescr.comboBoxItemsValues, True))
		mapColDefsCBlist.Put(oneColDescr.qualifiedColumnName.ToLowerCase, cbLst)
		' Insert or ignore
		If oneColDescr.unqualifiedColumnName.EqualsIgnoreCase(uniqueIDcolumnName) Then oneColDescr.allowEdit = 0 ' This must be an AUTOINCREMENT column, so don't allow editing it
		sqlMeta.ExecNonQuery2(s, Array(tablename, _
										oneColDescr.qualifiedColumnName, _
										oneColDescr.unqualifiedColumnName, _
										oneColDescr.columnHeader, _
										-1, _
										oneColDescr.alignment, _
										"", _
										-1, _
										oneColDescr.dataType, _
										0, _
										"", _
										"", _
										oneColDescr.allowEdit, _
										oneColDescr.allowEmpty, _
										oneColDescr.allowZero))
		' Update
		sqlMeta.ExecNonQuery2(s2, Array(oneColDescr.columnHeader, _
										oneColDescr.alignment, _
										oneColDescr.dataType, _
										oneColDescr.allowEdit, _
										oneColDescr.allowEmpty, _
										oneColDescr.allowZero, _
										tablename, _
										oneColDescr.qualifiedColumnName))
	Next

	' Renumber the column order in `TableViewColumns`, in case new columns were added from columnDefinitions
	colNum = -1
	rs = sqlMeta.ExecQuery2("SELECT * FROM `TableViewColumns` WHERE `TableName`=? ORDER BY `Position` ASC", Array As String(tablename))
	Do While rs.NextRow
		colNum = colNum + 1
		sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Position`=? WHERE `TableName`=? AND `QualifiedColumnName`=?", _
								Array(colNum, tablename, rs.GetString("QualifiedColumnName")))
	Loop
	rs.Close

End Sub

' Add a FontAwesome character and its colour for a specific value in the database table.
' This method can be called after Initialize and before Initialize2 is called.
' All arguments are case-insensitive.
' This is then used in the TableView in columns with a DataTypes.xxx_WITH_FONTAWESOME data type.
Public Sub AddFontAwesome(unqualifiedColumnName As String, dbValue As String, FAchar As String, intColour As Int)

	Dim m As Map
	m.Initialize
	m.Put("text", FAchar)
	m.Put("colour", fx.Colors.From32Bit(intColour))
	fontAwesomeMap.Put(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, m)

End Sub

' Add a MaterialIcons character and its colour for a specific value in the database table.
' This method can be called after Initialize and before Initialize2 is called.
' All arguments are case-insensitive.
' This is then used in the TableView in columns with a DataTypes.xxx_WITH_MATERIALICONS data type.
Public Sub AddMaterialIcon(unqualifiedColumnName As String, dbValue As String, MIchar As String, intColour As Int)

	Dim m As Map
	m.Initialize
	m.Put("text", MIchar)
	m.Put("colour", fx.Colors.From32Bit(intColour))
	materialIconsMap.Put(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, m)

End Sub

Private Sub GetIconMapText(unqualifiedColumnName As String, dbValue As String, fontAwesome As Boolean) As String

	Dim m As Map = IIf(fontAwesome, _
						fontAwesomeMap.GetDefault(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, Null), _
						materialIconsMap.GetDefault(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, Null))

	If m.IsInitialized Then
		If m <> Null Then Return m.GetDefault("text", "")
	End If

	Return ""

End Sub

Private Sub GetIconMapColour(unqualifiedColumnName As String, dbValue As String, fontAwesome As Boolean) As Paint

	Dim m As Map = IIf(fontAwesome, _
						fontAwesomeMap.GetDefault(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, Null), _
						materialIconsMap.GetDefault(unqualifiedColumnName.ToLowerCase & CRLF & dbValue.ToLowerCase, Null))

	If m.IsInitialized Then
		If m <> Null Then Return m.GetDefault("colour", fx.Colors.Black)
	End If

	Return fx.Colors.Black

End Sub

Private Sub Localise(s As String) As String

	Dim result As String = LocalisedStrings.GetDefault(s, "")
	Return IIf(result = "", s, result)

End Sub
#End Region ' Initalisations and settings

#Region Data changes
Private Sub DatabaseChanged(changeType As String)

	dbWasChanged = True

	If dbChangedCallbackSub <> "" Then CallSub3(caller, dbChangedCallbackSub, Me, changeType)

End Sub

Private Sub DeleteRows

	Dim listSelectedRows As List = TableViewGetSelectedRows(wmCRUDTableView1)
	Dim listSelectedIDs As List = TV1GetSelectedIDs(listSelectedRows, False)
	Dim row() As Object
	Dim lbl As Label
	Dim lblTag As wmCRUDlabelTagType
	Dim i As Int

	If fx.Msgbox2(MainForm, Localise("OK to delete?"), Localise("Please confirm"), Localise("Yes"), "", Localise("No"), fx.MSGBOX_CONFIRMATION) <> fx.DialogResponse.POSITIVE Then Return

	sql1.BeginTransaction
	For i = 0 To (listSelectedRows.Size - 1)
		row = listSelectedRows.Get(i)
		lbl = TV1GetLabelFromRow(row(0))
		lblTag = lbl.Tag
		Try
			If cannotDeleteIDvalues.IndexOf(lblTag.ID) < 0 Then
				sql1.ExecNonQuery2("DELETE FROM `" & tablename & "` WHERE `" & uniqueIDcolumnName & "`=?", Array As Object(lblTag.ID))
			End If
		Catch
			fx.Msgbox(MainForm, LastException, Localise("Error"))
			sql1.Rollback
			Return
		End Try
	Next
	sql1.TransactionSuccessful

	For i = (wmCRUDTableView1.Items.Size - 1) To 0 Step -1
		row = wmCRUDTableView1.Items.Get(i)
		lbl = TV1GetLabelFromRow(row(0))
		lblTag = lbl.Tag
		If listSelectedIDs.IndexOf(lblTag.ID) < 0 Then Continue
		wmCRUDTableView1.Items.RemoveAt(i)
	Next

	DatabaseChanged("d")
	wmCRUDTableView1.SelectedRow = -1

End Sub

Private Sub DuplicateRows

	Dim listSelectedRows As List = TableViewGetSelectedRows(wmCRUDTableView1)
	Dim listChangedIDs As List
	Dim i As Int
	Dim firstRow As Boolean = True
	Dim firstColumn As Boolean = True
	Dim lstColumnNames As List
	Dim columnNames As String = ""
	Dim valuePlaceHolders As String = ""
	Dim values As List

	listChangedIDs.Initialize
	values.Initialize

	sql1.BeginTransaction

	For i = 0 To (listSelectedRows.Size - 1)
		Dim row() As Object = listSelectedRows.Get(i)
		Dim lbl As Label = TV1GetLabelFromRow(row(0))
		Dim lblTag As wmCRUDlabelTagType = lbl.Tag
		Try
			Dim rset As ResultSet = sql1.ExecQuery2("SELECT * FROM `" & tablename & "` WHERE `" & uniqueIDcolumnName & "`=?", Array As Object(lblTag.ID))
			Do While rset.NextRow
				If firstRow Then
					lstColumnNames = SQLgetTableColumnNames(sql1, tablename, False)
					firstRow = False
					For Each oneColumnName As String In lstColumnNames
						If oneColumnName.EqualsIgnoreCase(uniqueIDcolumnName) Then Continue
						If firstColumn = False Then
							columnNames = columnNames & ", "
							valuePlaceHolders = valuePlaceHolders & ", "
						End If
						valuePlaceHolders = valuePlaceHolders & "?"
						columnNames = columnNames & "`" & oneColumnName & "`"
						firstColumn = False
					Next
				End If
				For Each oneColumnName As String In lstColumnNames
					If oneColumnName.EqualsIgnoreCase(uniqueIDcolumnName) Then Continue
					values.Add(rset.GetString(oneColumnName))
				Next
				sql1.ExecNonQuery2("INSERT INTO `" & tablename & "` (" & columnNames & ") VALUES (" & valuePlaceHolders & ")", values)
				GetLastAutoincValue(listChangedIDs)
				Exit
			Loop
			rset.close
		Catch
			fx.Msgbox(MainForm, LastException, Localise("Error"))
			sql1.Rollback
			rset.close
			Return
		End Try
	Next

	sql1.TransactionSuccessful
	DatabaseChanged("c")

	ShowData(listChangedIDs, True)

End Sub

Private Sub GetLastAutoincValue(listChangedIDs As List)

	Dim num As Int = -1

	Try
		' This SELECT only works for SQLite, not for other RDBMSs.
		' The only consequence for this class is that the TableView won't scroll to newly added items.
		num = sql1.ExecQuerySingleResult2("Select seq from sqlite_sequence where name=? COLLATE NOCASE", Array As String(tablename))
		If listChangedIDs.IsInitialized = False Then listChangedIDs.Initialize
		listChangedIDs.Add(num)
	Catch
		Log("") ' Just avoiding a compiler warning here
	End Try

End Sub

Private Sub Update_Or_Add(isAdd As Boolean)

	Dim s As String
	Dim i As Int
	Dim i2 As Int
	Dim row() As Object = IIf(isAdd, Null, wmCRUDTableView1.Items.Get(wmCRUDTableView1.SelectedRow))
	Dim lbl As Label = IIf(isAdd, Null, TV1GetLabelFromRow(row(0)))
	Dim lblTag As wmCRUDlabelTagType = IIf(isAdd, Null, lbl.Tag)
	Dim tagID As Int = IIf(isAdd, -1, lblTag.ID)
	Dim firstView As Boolean = True
	Dim firstNode As Node
	Dim allColumnLabels As List
	Dim allColumnViews As List
	Dim sqlInsert As String = ""
	Dim sqlInsertPlaceHolders As String = ""
	Dim sqlUpdate As String = ""
	Dim sqlValues As List
	Dim nextTop As Double = 5dip
	Dim viewWidth As Double
	Dim scrollPane1 As ScrollPane
	Dim listChangedIDs As List
	Dim lblError As Label

	SetWaitCursor(True)
	Sleep(50)

	TV1SaveColumnWidthsAndOrder(True)

	fillingDialog = True
	allColumnViews.Initialize
	allColumnLabels.Initialize
	sqlValues.Initialize
	update_add_dlg.Title = IIf(isAdd, Localise("New"), Localise("Edit"))

	' Setup the views for the columns
	Dim p As B4XView = IIf(editPane.IsInitialized, editPane, xui.CreatePanel(""))
	If editPane.IsInitialized Then
		For Each oneEditPaneView As Node In mapEditPaneViews.Values
			allColumnViews.Add(oneEditPaneView)
			If firstView Then firstNode = oneEditPaneView
			firstView = False
		Next
	Else
		p.SetLayoutAnimated(0, 0, 0, dialogWidth, dialogHeight)
		p.LoadLayout("wmcrud_add_update")
		scrollPane1.Initialize("")
		wmCRUDPaneAddUpdate.AddNode(scrollPane1, 0, 0, dialogWidth, dialogHeight)
		scrollPane1.SetVScrollVisibility("ALWAYS")
		scrollPane1.SetHScrollVisibility("NEVER")
		Dim p2 As Pane
		p2.Initialize("")
		scrollPane1.InnerNode = p2
		scrollPane1.InnerNode.SetSize(dialogWidth, dialogHeight)
		FormatPane(wmCRUDPaneAddUpdate)
		FormatPane(p2)
		Sleep(0) ' Let the UI update
		' Setup all columns' views in scrollPane1
		viewWidth = (p2.As(B4XView).Width - 30dip) / 2 ' Leave 5 dip at the front and back, and between label and data node, and keep 15dip for the vertical scrollbar
		For i = 0 To (columnDefinitionsOrderedLikeTV.Size - 1)
			Dim colDef As wmCRUDcolumnDefinition = columnDefinitionsOrderedLikeTV.Get(i).As(wmCRUDcolumnDefinition)
			Dim n As Node = Update_Or_Add_CreateView(p2, nextTop, colDef, viewWidth, allColumnViews, allColumnLabels)
			If firstView And (colDef.allowEdit <> 0) Then
				firstNode = n
				firstView = False
			End If
			nextTop = nextTop + n.As(B4XView).Height + 5dip
		Next
		' Add the label for error messages (from the caller, set in Sub SetValidationError) at the bottom
		lblError.Initialize("")
		lblError.SetLayoutAnimated(0, 5dip, nextTop, p2.As(B4XView).Width - 10dip, filterFieldHeight)
		lblError.Text = ""
		lblError.Alignment = "CENTER_LEFT"
		FormatLabel(lblError)
		lblError.As(B4XView).TextColor = xui.Color_Red
		nextTop = Max(nextTop, p2.As(B4XView).Height - 5dip - filterFieldHeight)
		p2.AddNode(lblError, 5dip, nextTop, p2.As(B4XView).Width - 10dip, filterFieldHeight)
	End If

	' Fill all columns' views (when updating) and prepare the sql... variables
	For i = 0 To (columnDefinitionsOrderedLikeTV.Size - 1)
		Dim colDef As wmCRUDcolumnDefinition = columnDefinitionsOrderedLikeTV.Get(i).As(wmCRUDcolumnDefinition)
		Update_Or_Add_FillView(i, isAdd, allColumnViews.Get(i).As(Node), colDef, row)
		If colDef.unqualifiedColumnName.EqualsIgnoreCase(uniqueIDcolumnName) Then Continue ' This must be an AUTOINCREMENT column, so don't add it to the data to add/update
		If sqlInsert <> "" Then sqlInsert = sqlInsert & ", "
		sqlInsert = sqlInsert & "`" & colDef.unqualifiedColumnName & "`"
		If sqlInsertPlaceHolders <> "" Then sqlInsertPlaceHolders = sqlInsertPlaceHolders & ", "
		sqlInsertPlaceHolders = sqlInsertPlaceHolders & "?"
		If sqlUpdate <> "" Then sqlUpdate = sqlUpdate & ", "
		sqlUpdate = sqlUpdate & "`" & colDef.unqualifiedColumnName & "`=?"
	Next

	Sleep(0) ' Let the UI update

	If isAdd Then
		Dim rSub As ResumableSub = update_add_dlg.ShowCustom(p, Localise("Save"), "", Localise("Cancel"))
	Else If allowNew Then
		Dim rSub As ResumableSub = update_add_dlg.ShowCustom(p, Localise("Save"), Localise("Save new"), Localise("Cancel"))
	Else
		Dim rSub As ResumableSub = update_add_dlg.ShowCustom(p, Localise("Save"), "", Localise("Cancel"))
	End If

	If firstNode.IsInitialized Then firstNode.RequestFocus

	If dataValidationCallbackSub <> "" Then
		updateOrAddDataValidation.Initialize
		updateOrAddDataValidation.instance = Me
		updateOrAddDataValidation.dlg = update_add_dlg
		updateOrAddDataValidation.columnViews = allColumnViews
		updateOrAddDataValidation.columnLabels = allColumnLabels
		updateOrAddDataValidation.btnCancel = update_add_dlg.GetButton(xui.DialogResponse_Cancel)
		updateOrAddDataValidation.btnSaveNew = update_add_dlg.GetButton(xui.DialogResponse_Negative)
		updateOrAddDataValidation.btnSave = update_add_dlg.GetButton(xui.DialogResponse_Positive)
		updateOrAddDataValidation.errorLabel = lblError
		Update_Or_Add_CallerDataValidation
	End If

	fillingDialog = False

	SetWaitCursor(False)
	Sleep(0)

	Wait For (rSub) Complete(Result As Int)
	update_add_dlg.PutAtTop = False
	Select Case Result
		Case xui.DialogResponse_Cancel
			Return
		Case xui.DialogResponse_Negative
			isAdd = True ' An existing row was double-clicked; save the edited version as a new row
	End Select

	For Each n As Node In allColumnViews
		If n.Id.EqualsIgnoreCase(uniqueIDcolumnName) Then Continue ' This must be an AUTOINCREMENT column, so don't add it to the data to add/update
		Dim colDefValue As wmCRUDcolumnDefinition
		For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitions
			If oneColDef.unqualifiedColumnName = n.Id Then
				colDefValue = oneColDef
				Exit
			End If
		Next
		Dim nodeDataType As String = mapDataTypes.GetDefault(colDefValue.dataType, "t")
		If n Is ComboBox Then
			Dim cb As ComboBox = n
			sqlValues.Add(GetComboboxDbValue(colDefValue.qualifiedColumnName, cb.Value.As(String)))
		Else If n Is DatePicker Then
			Dim dp As DatePicker = n
			If dp.DateTicks = 0 Then
				sqlValues.Add("")
			Else
				Select Case nodeDataType
					Case "dtymd" ' YYYYMMDD
						sqlValues.Add(DateToYYYYMMDD(dp.DateTicks, ""))
					Case "dtlong"
						sqlValues.Add(dp.DateTicks)
					Case "dtymdh" ' YYYY-MM-DD
						sqlValues.Add(DateToYYYYMMDD(dp.DateTicks, "-"))
					Case Else
						sqlValues.Add("WMCRUD CODING ERROR")
						Log("wmCRUD.Update_Or_Add (" & tablename & "): unhandled date datatype (coding oversight) for column: " & colDefValue.qualifiedColumnName)
				End Select
			End If
		Else If n Is TextArea Then
			Dim ta As TextArea = n
			sqlValues.Add(ta.Text)
		Else If n Is TextField Then
			Dim tf As TextField = n
			Select Case nodeDataType
				Case "i"
					Try
						s = tf.Text.Replace(thousandsSeparator, "").As(Int)
						Dim iVal As Int = s
						sqlValues.Add(iVal)
					Catch
						sqlValues.Add(s)
						Log("wmCRUD.Update_Or_Add (" & tablename & "): invalid integer value ('" & s & "') not caught by data validation for: " & colDefValue.qualifiedColumnName)
					End Try
				Case "d"
					Try
						s = tf.Text.Replace(thousandsSeparator, "").As(Double)
						Dim dVal As Double = s
						sqlValues.Add(dVal)
					Catch
						sqlValues.Add(s)
						Log("wmCRUD.Update_Or_Add (" & tablename & "): invalid double value ('" & s & "') not caught by data validation for: " & colDefValue.qualifiedColumnName)
					End Try
				Case Else
					sqlValues.Add(tf.Text)
			End Select
		Else
			sqlValues.Add("WMCRUD CODING ERROR")
			Log("wmCRUD.Update_Or_Add (" & tablename & "): unhandled view type (coding oversight) for column: " & colDefValue.qualifiedColumnName)
		End If
	Next

	If isAdd Then
		Try
			sql1.ExecNonQuery2("INSERT INTO `" & tablename & "` (" & sqlInsert & ") VALUES (" & sqlInsertPlaceHolders & ")", sqlValues)
			GetLastAutoincValue(listChangedIDs)
		Catch
			fx.Msgbox(MainForm, LastException, Localise("Error"))
			Return
		End Try
	Else
		' Update the TableView
		i2 = -1
		For i = 0 To (columnDefinitionsOrderedLikeTV.Size - 1)
			colDef = columnDefinitionsOrderedLikeTV.Get(i).as(wmCRUDcolumnDefinition)
			If colDef.unqualifiedColumnName.EqualsIgnoreCase(uniqueIDcolumnName) Then Continue ' This must be an AUTOINCREMENT column, so its value was not added to sqlValues
			i2 = i2 + 1
			UpdateBackToTableView(row, i2, colDef, sqlValues.Get(i2).As(String))
		Next
		' Update the database
		sqlValues.Add(tagID)
		Try
			sql1.ExecNonQuery2("UPDATE `" & tablename & "` SET " & sqlUpdate & " WHERE `" & uniqueIDcolumnName & "`=?", sqlValues)
			listChangedIDs.Initialize
			listChangedIDs.Add(tagID)
		Catch
			fx.Msgbox(MainForm, LastException, Localise("Error"))
			Return
		End Try
	End If

	DatabaseChanged(IIf(isAdd, "c", "u"))
	ShowData(listChangedIDs, isAdd)

End Sub

Private Sub Update_Or_Add_CreateView(innernode As Pane, nextTop As Double, colDef As wmCRUDcolumnDefinition, viewWidth As Double, allColumnViews As List, allColumnLabels As List) As Node

	Dim n As Node
	Dim lbl As Label

	lbl.Initialize("")
	lbl.SetLayoutAnimated(0, 5dip, nextTop, viewWidth, filterFieldHeight)
	lbl.Text = colDef.columnHeader
	lbl.Alignment = "CENTER_RIGHT"
	FormatLabel(lbl)
	lbl.As(B4XView).TextColor = clr_foregroundInt
	innernode.AddNode(lbl, 5dip, nextTop, viewWidth, lbl.As(B4XView).Height)
	allColumnLabels.Add(lbl)

	If colDef.dataType = WMCRUD_DATATYPE_COMBOBOX Then
		Dim cb As ComboBox
		cb.Initialize("wmCRUD_UAcomboBox")
		cb.As(B4XView).Height = filterFieldHeight
		n = cb
	Else If (colDef.dataType > WMCRUD_DATATYPE_DATE_START) And (colDef.dataType < WMCRUD_DATATYPE_DATE_END) Then
		Dim dp As DatePicker
		dp.Initialize("wmCRUD_UAdatePicker")
		dp.PickOnBounds = True
		dp.Editable = False
		dp.As(B4XView).Height = filterFieldHeight
		n = dp
	Else If colDef.dataType = WMCRUD_DATATYPE_MULTILINE_TEXT Then
		Dim ta As TextArea
		ta.Initialize("wmCRUD_UAtextArea")
		ta.As(B4XView).Height = 5 * filterFieldHeight
		n = ta
	Else
		Dim tf As TextField
		tf.Initialize("wmCRUD_UAtextField")
		tf.PromptText = ""
		tf.As(B4XView).Height = filterFieldHeight
		n = tf
	End If

	n.id = colDef.unqualifiedColumnName
	n.Tag = colDef.qualifiedColumnName
	n.PickOnBounds = True
	innernode.AddNode(n, viewWidth + 10dip, nextTop, viewWidth, n.As(B4XView).Height)
	allColumnViews.Add(n)

	Return n

End Sub

Private Sub Update_Or_Add_FillView(i As Int, isAdd As Boolean, n As Node, colDef As wmCRUDcolumnDefinition, row() As Object)

	Dim labelInRow As Label = TV1GetLabelFromRow(row(i))
	Dim cellVal As String = IIf(isAdd, "", labelInRow.Text)

	n.Enabled = (colDef.allowEdit <> 0)

	If n Is ComboBox Then
		Dim cb As ComboBox = n
		cb.Items.Clear
		Dim cbLst As List = GetComboboxLists(cb.Tag.As(String))
		Dim cbLstDisplay As List = cbLst.Get(0).As(List)
		cb.Items.AddAll(cbLstDisplay)
		If isAdd Then
			cb.SelectedIndex = -1
		Else
			If cbLstDisplay.IndexOf(cellVal) >= 0 Then cb.Value = cellVal
		End If
		If editPane.IsInitialized = False Then FormatComboBox(cb)
	Else If n Is DatePicker Then
		Dim dp As DatePicker = n
		If isAdd Then
			dp.DateTicks = 0
		Else If cellVal = "" Then
			dp.DateTicks = 0
		Else ' Cellval is DD/MM/YYYY
			dp.DateTicks = DDMMYYYYtoDate(cellVal, "/")
		End If
	Else If n Is TextArea Then
		Dim ta As TextArea = n
		If isAdd Then
			ta.Text = ""
		Else
			ta.Text = cellVal
		End If
		If editPane.IsInitialized = False Then FormatTextArea(ta)
	Else If n Is TextField Then
		Dim tf As TextField = n
		If isAdd Then
			tf.Text = ""
		Else If (colDef.dataType > WMCRUD_DATATYPE_WITH_FONTAWESOME_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_FONTAWESOME_END) Then
			tf.Text = labelInRow.TooltipText
		Else If (colDef.dataType > WMCRUD_DATATYPE_WITH_MATERIALICONS_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_MATERIALICONS_END) Then
			tf.Text = labelInRow.TooltipText
		Else
			Select Case mapDataTypes.GetDefault(colDef.dataType, "t")
				Case "i"
					Try
						Dim iVal As Int = cellVal.As(Int)
						tf.Text = NumberFormat2(iVal, 1, 0, 0, True)
					Catch
						tf.Text = cellVal
					End Try
				Case "d"
					Try
						Dim dVal As Double = cellVal.As(Double)
						tf.Text = NumberFormat2(dVal, 1, 2, 2, True)
					Catch
						tf.Text = cellVal
					End Try
				Case Else
					tf.Text = cellVal
			End Select
		End If
		If editPane.IsInitialized = False Then FormatTextField(tf)
	Else
		Log("wmCRUD.Update_Or_Add_FillView (" & tablename & "): unhandled view type (coding oversight) for column: " & colDef.qualifiedColumnName)
	End If

End Sub

Private Sub FormatComboBox(cb As ComboBox)

	cb.As(B4XView).Font = arialGeneral.As(B4XFont)
	cb.As(B4XView).SetColorAndBorder(clr_backgroundInt, borderWidth, clr_foregroundInt, borderRadius)

End Sub

Private Sub FormatLabel(lbl As B4XView)

	lbl.As(B4XView).Font = arialGeneral.As(B4XFont)

End Sub

Private Sub FormatPane(p As B4XView)

	p.Color = clr_backgroundInt

End Sub

Private Sub FormatTextField(tf As B4XView)

	tf.As(B4XView).Font = arialGeneral.As(B4XFont)
	tf.TextColor = clr_foregroundInt
	tf.Color = clr_backgroundInt
	CSSUtils.SetBorder(tf, borderWidth, clr_foregroundPaint, borderRadius)

End Sub

Private Sub FormatTextArea(v As TextArea)

	' Do this to avoid TextAreas getting a blue background when they receive focus:
	' https://www.b4x.com/android/forum/threads/textarea-transparent-background-color-on-focus.141893/#post-899334
	CSSUtils.SetStyleProperty(v,"-fx-control-inner-background","transparent")
	CSSUtils.SetStyleProperty(v,"-fx-background","transparent")
	CSSUtils.SetStyleProperty(v,"-fx-faint-focus-color","transparent")
	CSSUtils.SetStyleProperty(v,"-fx-background-color","transparent")

	v.As(B4XView).Font = arialGeneral.As(B4XFont)

	CSSUtils.SetStyleProperty(v,"-fx-text-fill", ColorToHexCSS(clr_foregroundInt))
	CSSUtils.SetBorder(v, borderWidth, clr_foregroundPaint, borderRadius)

End Sub

' This one must be called by all DatePickers' FocusChanged event method
Private Sub DatePicker_FocusChanged(dp As DatePicker, HasFocus As Boolean)

	' Code based on https://www.b4x.com/android/forum/threads/datepicker-behaviour-differs-between-oracle-java-8-and-openjdk-11.132956/

	If HasFocus = False Then
		Dim s As String = dp.As(JavaObject).RunMethodJO("getEditor", Null).RunMethod("getText", Null)
		Try
			Dim date As Long = DateTime.DateParse(s)
			dp.DateTicks = date
		Catch
			dp.As(JavaObject).RunMethodJO("getEditor", Null).RunMethod("setText", Array(DateTime.Date(dp.DateTicks)))
		End Try
	End If

End Sub

Private Sub UpdateBackToTableView(row() As Object, i As Int, colDef As wmCRUDcolumnDefinition, cellVal As String)

	Dim dataType As String = mapDataTypes.GetDefault(colDef.dataType, "t")
	Dim lbl As Label
	Dim v As B4XView = row(i)
	If v Is Pane Then
		lbl = v.GetView(0)
	Else
		lbl = row(i)
	End If

	If (colDef.dataType > WMCRUD_DATATYPE_WITH_FONTAWESOME_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_FONTAWESOME_END) Then
		lbl.Text = GetIconMapText(colDef.unqualifiedColumnName, cellVal, True)
		lbl.TextColor = GetIconMapColour(colDef.unqualifiedColumnName, cellVal, True)
	Else If (colDef.dataType > WMCRUD_DATATYPE_WITH_MATERIALICONS_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_MATERIALICONS_END) Then
		lbl.Text = GetIconMapText(colDef.unqualifiedColumnName, cellVal, False)
		lbl.TextColor = GetIconMapColour(colDef.unqualifiedColumnName, cellVal, False)
	Else
		Select Case dataType
			Case "cb"
				lbl.Text = GetComboboxDisplayValue(colDef.qualifiedColumnName, cellVal)
			Case "i"
				Try
					Dim iVal As Int = cellVal.As(Int)
					lbl.Text = NumberFormat2(iVal, 1, 0, 0, True)
				Catch
					lbl.Text = cellVal
				End Try
			Case "d"
				Try
					Dim dVal As Double = cellVal.As(Double)
					lbl.Text = NumberFormat2(dVal, 1, 2, 2, True)
				Catch
					lbl.Text = cellVal
				End Try
			Case "dtymd"
				Try
					lbl.Text = cellVal.SubString(6) & "/" & cellVal.SubString2(4, 6) & "/" & cellVal.SubString2(0, 4) ' Show YYYYMMDD as DD/MM/YYYY
				Catch
					lbl.Text = cellVal
				End Try
			Case "dtymdh"
				Try
					lbl.Text = cellVal.SubString(8) & "/" & cellVal.SubString2(5, 7) & "/" & cellVal.SubString2(0, 4) ' Show YYYY-MM-DD as DD/MM/YYYY
				Catch
					lbl.Text = cellVal
				End Try
			Case "dtlong"
				Try
					Dim lVal As Long = cellVal.As(Long)
					If lVal > 0 Then lbl.Text = DateTime.Date(lVal) ' Assuming that '0' means 'no date'
				Catch
					lbl.Text = cellVal
				End Try
			Case Else ' "t", "tm", "cb"
				lbl.Text = cellVal
		End Select
	End If

End Sub

Private Sub wmCRUD_UAcomboBox_ValueChanged(Value As Object)

	If fillingDialog = False Then Update_Or_Add_CallerDataValidation

End Sub

Private Sub wmCRUD_UAdatePicker_FocusChanged(HasFocus As Boolean)

	Dim dp As DatePicker = Sender
	DatePicker_FocusChanged(dp, HasFocus)

End Sub

Private Sub wmCRUD_UAdatePicker_ValueChanged(Value As Long)

	If fillingDialog = False Then Update_Or_Add_CallerDataValidation

End Sub

Private Sub wmCRUD_UAtextArea_TextChanged(Old As String, New As String)

	If fillingDialog = False Then Update_Or_Add_CallerDataValidation

End Sub

Private Sub wmCRUD_UAtextField_TextChanged(Old As String, New As String)

	Dim tf As TextField = Sender
	Dim newNoThousands As String = New.Replace(thousandsSeparator, "")

	If fillingDialog Then Return

	updateOrAddDataValidation.errorLabel.Text = ""

	' Already catch/remove illegal characters and check for empty/zero values
	For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitionsOrderedLikeTV
		If oneColDef.unqualifiedColumnName <> tf.id Then Continue
		Dim tfLabel As Label = FindLabelInUpdateOrAddDataValidation(tf.Id)
		tfLabel.TextColor = clr_foregroundPaint
		If (New = "") And (oneColDef.allowEmpty = 0) Then
			wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "A value is required here")
			Return
		End If
		If IsNumber(New) And (oneColDef.allowZero = 0) And (New.As(Double) = 0) Then
			wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "A non-zero value is required here")
			Return
		End If
		If tf.Text = "" Then Exit
		Select Case oneColDef.dataType
			Case WMCRUD_DATATYPE_INT, WMCRUD_DATATYPE_INT_WITH_FONTAWESOME, WMCRUD_DATATYPE_INT_WITH_MATERIALICONS
				If IsSignedInteger(newNoThousands, True) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be an integer number")
					Return
				End If
			Case WMCRUD_DATATYPE_INT_POS, WMCRUD_DATATYPE_INT_POS_WITH_FONTAWESOME, WMCRUD_DATATYPE_INT_WITH_MATERIALICONS
				If IsSignedInteger(newNoThousands, True) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a positive integer number")
					Return
				Else
					Dim iVal As Int = newNoThousands.As(Int)
					If iVal < 0 Then
						wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a positive integer number")
						Return
					End If
				End If
			Case WMCRUD_DATATYPE_INT_NEG, WMCRUD_DATATYPE_INT_NEG_WITH_FONTAWESOME, WMCRUD_DATATYPE_INT_NEG_WITH_MATERIALICONS
				If IsSignedInteger(newNoThousands, True) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a negative integer number")
					Return
				Else
					Dim iVal As Int = newNoThousands.As(Int)
					If iVal > 0 Then
						wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a negative integer number")
						Return
					End If
				End If
			Case WMCRUD_DATATYPE_DEC, WMCRUD_DATATYPE_DEC_WITH_FONTAWESOME, WMCRUD_DATATYPE_DEC_WITH_MATERIALICONS
				If IsSignedDecimal(newNoThousands, True, decimalSeparator) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a decimal number")
					Return
				End If
			Case WMCRUD_DATATYPE_DEC_POS, WMCRUD_DATATYPE_DEC_POS_WITH_FONTAWESOME, WMCRUD_DATATYPE_DEC_POS_WITH_MATERIALICONS
				If IsSignedDecimal(newNoThousands, True, decimalSeparator) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a positive decimal number")
					Return
				Else
					Dim dVal As Double = newNoThousands.As(Double)
					If dVal < 0 Then
						wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a positive decimal number")
						Return
					End If
				End If
			Case WMCRUD_DATATYPE_DEC_NEG, WMCRUD_DATATYPE_DEC_NEG_WITH_FONTAWESOME, WMCRUD_DATATYPE_DEC_NEG_WITH_MATERIALICONS
				If IsSignedDecimal(newNoThousands, True, decimalSeparator) = False Then
					wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a negative decimal number")
					Return
				Else
					Dim dVal As Double = newNoThousands.As(Double)
					If dVal > 0 Then
						wmCRUD_UAtextField_TextChanged_Error(tf, tfLabel, "This must be a negative decimal number")
						Return
					End If
				End If
		End Select
		Exit
	Next

	Update_Or_Add_CallerDataValidation

End Sub

Private Sub wmCRUD_UAtextField_TextChanged_Error(tf As TextField, tfLabel As Label, errMsg As String)

	updateOrAddDataValidation.btnSave.Enabled = False
	If updateOrAddDataValidation.btnSaveNew.IsInitialized Then updateOrAddDataValidation.btnSaveNew.Enabled = False ' The 'Save new' button is not always present, hence the test for IsInitialized
	SetValidationError(tf, Localise(errMsg) & " (" & tfLabel.Text & ")")

End Sub

Private Sub Update_Or_Add_CallerDataValidation

	If dataValidationCallbackSub = "" Then
		updateOrAddDataValidation.btnSave.Enabled = True
		If updateOrAddDataValidation.btnSaveNew.IsInitialized Then updateOrAddDataValidation.btnSaveNew.Enabled = True ' The 'Save new' button is not always present, hence the test for IsInitialized
	Else
		CallSub2(caller, dataValidationCallbackSub, updateOrAddDataValidation)
	End If

End Sub

' Sets red text for a view's label in the update/add dialogue, and adds an error message.
Public Sub SetValidationError(n As Node, errMsg As String)

	updateOrAddDataValidation.errorLabel.Text = errMsg
	FindLabelInUpdateOrAddDataValidation(n.Id).TextColor = fx.Colors.Red

End Sub

' Resets the red text for a view's label in the update/add dialogue, and removes any error message.
Public Sub ClearValidationError(n As Node)

	updateOrAddDataValidation.errorLabel.Text = ""
	FindLabelInUpdateOrAddDataValidation(n.Id).TextColor = clr_foregroundPaint

End Sub

Private Sub FindLabelInUpdateOrAddDataValidation(nodeId As String) As Label

	Dim i As Int

	For i = 0 To (updateOrAddDataValidation.columnViews.Size - 1)
		Dim v As Node = updateOrAddDataValidation.columnViews.Get(i)
		If v.Id = nodeId Then Return updateOrAddDataValidation.columnLabels.Get(i)
	Next

	Return Null

End Sub
#End Region ' Data changes

#Region Show Data
' Shows the data.
Public Sub Show

	If isInitialized2 = False Then
		Log("wmCRUD.Show - Initialize2 has not been called")
		Return
	End If

	FormatPane(wmCRUDPaneStatus)

	wmCRUDLabelBack.As(B4XView).TextColor = clr_foregroundInt
	wmCRUDLabelBack.As(B4XView).Color = clr_backgroundInt
	SetToolTip(wmCRUDLabelBack, Localise("Back"), arialBoldGeneral, True)

	wmCRUDLabelRefresh.As(B4XView).TextColor = clr_foregroundInt
	wmCRUDLabelRefresh.As(B4XView).Color = clr_backgroundInt
	SetToolTip(wmCRUDLabelRefresh, Localise("Refresh"), arialBoldGeneral, True)

	FormatLabel(wmCRUDLabelStatus)
	wmCRUDLabelStatus.As(B4XView).TextColor = clr_foregroundInt
	wmCRUDLabelStatus.As(B4XView).Color = clr_backgroundInt

	FormatPane(wmCRUDPaneFilter)
	wmCRUDPaneFilter.Visible = myShowFilters
	wmCRUDTableView1.Left = IIf(myShowFilters, TV1originalLeft, wmCRUDPaneFilter.Left)

	ShowData(Null, False)

End Sub

Private Sub ShowData(listChangedIDs As List, wasAdded As Boolean)

	Dim rs As ResultSet
	Dim orderClause As String = ""
	Dim rowWithTheChangedID As Int = -1
	Dim sqlQuery As String
	Dim i As Int
	Dim defaultColWidth As Double = 100dip
	Dim rowPos As Int
	Dim updatedID As Int = -1

	resizedColumnsTimer.Initialize("ResizedColumnsTimer", 0)
	resizedColumnsTimer.Enabled = False

	If (wasAdded = False) And (listChangedIDs.IsInitialized = False) Then
		TV1SaveColumnWidthsAndOrder(True)
		wmCRUDPaneFilter.RemoveAllNodes
	End If
	If sql1.IsInitialized = False Then Return

	SetWaitCursor(True)
	Sleep(50)

	showingData = True

	If listChangedIDs.IsInitialized = False Then wmCRUDTableView1.Items.Clear
	wmCRUDTableView1.Enabled = False
	wmCRUDTableView1.PickOnBounds = True

	Sleep(0)

	If listChangedIDs.IsInitialized = False Then
		If (defaultColWidth * columnDefinitions.Size) < wmCRUDTableView1.Width Then
			defaultColWidth = wmCRUDTableView1.Width / columnDefinitions.Size
		End If
		sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Width`=? WHERE `TableName`=? AND `Width`<0", Array As String(defaultColWidth, tablename))
		TV1SetColumns
	End If

	SetColumnDefinitionsOrderedLikeTV

	If dbSortOrder <> "" Then orderClause = " ORDER BY " & dbSortColumn & " " & dbSortOrder

	sqlQuery = "SELECT * FROM `" & tablename & "` "

	If dbWhereFilter = "" Then
		rs = sql1.ExecQuery(sqlQuery & orderClause)
	Else
		rs = sql1.ExecQuery2(sqlQuery & dbWhereFilter & orderClause, dbWhereFilterValuesList)
	End If

	' Only remove the updated row and then re-insert it afterwards.
	' While the TableView was updated in Update_Or_Add, the changed data
	' and sort order may influence the position of the row in the list.
	If listChangedIDs.IsInitialized And (wasAdded = False) Then
		For i = (wmCRUDTableView1.Items.Size - 1) To 0 Step -1
			Dim row2() As Object = wmCRUDTableView1.Items.Get(i)
			Dim row2lbl As Label = TV1GetLabelFromRow(row2(0))
			Dim row2lblTag As wmCRUDlabelTagType = row2lbl.Tag
			If listChangedIDs.IndexOf(row2lblTag.ID) >= 0 Then
				updatedID = row2lblTag.ID
				wmCRUDTableView1.Items.RemoveAt(i)
				Exit ' There is only one updated row
			End If
		Next
	End If

	rowPos = -1
	Do While rs.NextRow
		rowPos = rowPos + 1
		Dim dbValues As List
		dbValues.Initialize
		For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitionsOrderedLikeTV
			dbValues.Add(rs.GetString(oneColDef.unqualifiedColumnName))
		Next
		If updatedID >= 0 Then ' Just updating 1 row, which must be re-inserted
			If rs.GetInt(uniqueIDcolumnName) = updatedID Then
				rowWithTheChangedID = rowPos
				wmCRUDTableView1.Items.InsertAt(rowPos, TV1CreateRow(dbValues, rs.GetInt(uniqueIDcolumnName), columnDefinitionsOrderedLikeTV))
				Exit
			End If
		Else
			wmCRUDTableView1.Items.Add(TV1CreateRow(dbValues, rs.GetInt(uniqueIDcolumnName), columnDefinitionsOrderedLikeTV))
		End If
	Loop
	rs.Close

	ScrollToTableViewRow(listChangedIDs, rowWithTheChangedID)

	SetWaitCursor(False)
	Sleep(0)

	If wmCRUDTableView1.Items.Size > 0 Then
		wmCRUDLabelStatus.Visible = True
		wmCRUDLabelStatus.Text = IIf(tableDisplayName = "", "", "[" & tableDisplayName & "] ") & Localise("Rows showing") & ": " & wmCRUDTableView1.Items.Size
	Else
		wmCRUDLabelStatus.Visible = False
	End If
	wmCRUDTableView1.Enabled = True
	showingDataAfterSettingFilter = False
	showingData = False

End Sub

Private Sub ScrollToTableViewRow(listChangedIDs As List, rowWithTheChangedID As Int)

	Try
		If wmCRUDTableView1.Items.Size > 0 Then
			If listChangedIDs.IsInitialized And (rowWithTheChangedID >= 0) Then
				wmCRUDTableView1.ScrollTo(rowWithTheChangedID)
				wmCRUDTableView1.SelectedRow = rowWithTheChangedID
			Else
				wmCRUDTableView1.ScrollTo(wmCRUDTableView1.Items.Size - 1)
			End If
		End If
	Catch
		Log("wmCRUD.ScrollToTableViewRow (" & tablename & "): " & LastException)
	End Try

End Sub
#End Region ' Show Data

#Region TableView
Private Sub TV1CreateRow(dbValues As List, ID As Int, colDefs As List) As Object()

	Dim row(dbValues.Size) As Object
	Dim i As Int

	' When running *natively* on Linux, the FontAwesome characters turned into garbage
	' in the TableView if a straightforward label was added. I tracked this down to
	' "lbl.As(B4XView).Color = clr_backgroundInt" (for the text label) which caused
	' the problem, probably because CSSUtils is used to set 'Color' for a B4XView,
	' and I've seen a post on the forum about CSSUtils/additional CSS causing such
	' problems. Hence the workaround with putting the label in a pane; that works
	' on all platforms (Linux, Linux+wine, Windows).

	For i = 0 To (dbValues.Size - 1)
		Dim cellVal As String = NullOrEmptyStringToDefault(dbValues.Get(i).As(String), "")
		Dim colDef As wmCRUDcolumnDefinition = colDefs.Get(i)
		Dim lbl As Label
		Dim lblTag As wmCRUDlabelTagType
		lblTag.ID = ID
		If (colDef.dataType > WMCRUD_DATATYPE_WITH_FONTAWESOME_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_FONTAWESOME_END) Then
			lbl = TV1SetFAlabel(i, lbl, GetIconMapText(colDef.unqualifiedColumnName, cellVal, True), GetIconMapColour(colDef.unqualifiedColumnName, cellVal, True), cellVal)
			lbl.Tag = lblTag
			row(i) = lbl
		Else If (colDef.dataType > WMCRUD_DATATYPE_WITH_MATERIALICONS_START) And (colDef.dataType < WMCRUD_DATATYPE_WITH_MATERIALICONS_END) Then
			lbl = TV1SetMIlabel(i, lbl, GetIconMapText(colDef.unqualifiedColumnName, cellVal, False), GetIconMapColour(colDef.unqualifiedColumnName, cellVal, False), cellVal)
			lbl.Tag = lblTag
			row(i) = lbl
		Else
			Dim lblPane As B4XView = xui.CreatePanel("")
			lblPane.Color = clr_backgroundInt
			lblPane.Width = wmCRUDTableView1.GetColumnWidth(i)
			lblPane.Height = tvCellSize
			lbl.Initialize("TvLbl")
			lblPane.AddView(lbl, 0, 0, wmCRUDTableView1.GetColumnWidth(i), 30dip) ' The same value as "-fx-cell-size" in file 'wmcrud_tvformat.css'
			lbl.PickOnBounds = True
			Select Case mapDataTypes.GetDefault(colDef.dataType, "t")
				Case "cb"
					lbl.Text = GetComboboxDisplayValue(colDef.qualifiedColumnName, cellVal)
				Case "i"
					Try
						Dim iVal As Int = cellVal.As(Int)
						lbl.Text = NumberFormat2(iVal, 1, 0, 0, True)
					Catch
						lbl.Text = cellVal
					End Try
				Case "d"
					Try
						Dim dVal As Double = cellVal.As(Double)
						lbl.Text = NumberFormat2(dVal, 1, 2, 2, True)
					Catch
						lbl.Text = cellVal
					End Try
				Case "dtymd"
					Try
						lbl.Text = cellVal.SubString(6) & "/" & cellVal.SubString2(4, 6) & "/" & cellVal.SubString2(0, 4) ' Show YYYYMMDD as DD/MM/YYYY
					Catch
						lbl.Text = cellVal
					End Try
				Case "dtymdh"
					Try
						lbl.Text = cellVal.SubString(8) & "/" & cellVal.SubString2(5, 7) & "/" & cellVal.SubString2(0, 4) ' Show YYYY-MM-DD as DD/MM/YYYY
					Catch
						lbl.Text = cellVal
					End Try
				Case "dtlong"
					Try
						Dim lVal As Long = cellVal.As(Long)
						If lVal > 0 Then lbl.Text = DateTime.Date(lVal) ' Assuming that '0' means 'no date'
					Catch
						lbl.Text = cellVal
					End Try
				Case Else
					lbl.Text = cellVal
			End Select
			lbl.TextColor = clr_foregroundPaint
			Select Case colDef.alignment
				Case "CENTER"
					lbl.Alignment = "CENTER"
				Case "RIGHT"
					lbl.Alignment = "CENTER_RIGHT"
				Case Else
					lbl.Alignment = "CENTER_LEFT"
			End Select
			lbl.Tag = lblTag
			lblPane.Tag = lblTag
			row(i) = lblPane
		End If
	Next

	Return row

End Sub

Private Sub wmCRUDTableView1_MouseClicked(EventData As MouseEvent)

	If allowEdit = False Then Return
	If EventData.ClickCount <> 2 Then Return ' Not a double click
	If wmCRUDTableView1.SelectedRow < 0 Then Return ' Header row clicked
	If TableViewGetSelectedRows(wmCRUDTableView1).Size <> 1 Then Return ' Exactly 1 row should be selected for editing

	Dim row() As Object = wmCRUDTableView1.Items.Get(wmCRUDTableView1.SelectedRow)
	If row = Null Then Return

	Dim lbl As Label = TV1GetLabelFromRow(row(0))
	Dim lblTag As wmCRUDlabelTagType = lbl.Tag
	If cannotEditIDvalues.IndexOf(lblTag.ID) >= 0 Then Return ' Edit not allowed for this ID

	Update_Or_Add(False)

End Sub

Private Sub TV1SetColumns

	Dim rs As ResultSet
	Dim colNum As Int
	Dim colHdrList As List
	Dim filterNum As Int

	dbSortOrder = ""
	dbSortColumn = ""

	TV1SetColumnsInProgress = True

	' Filters
	If myShowFilters Then
		If showingDataAfterSettingFilter = False Then ResetDbWhereFilters(False)
		lblFilters.Initialize("LblFilters")
		lblFilters.PickOnBounds = True
		lblFilters.Font = arialBoldGeneral
		lblFilters.Text = Localise("Filters")
		lblFilters.As(B4XView).Color = clr_backgroundInt
		lblFilters.As(B4XView).TextColor = IIf(showingDataAfterSettingFilter, xui.Color_Red, clr_foregroundInt)
		lblFilters.TooltipText = Localise("Click to reset all filters")
		SetToolTipFont(lblFilters, arialBoldGeneral)
		lblFilters.Alignment = "CENTER"
		wmCRUDPaneFilter.AddNode(lblFilters, 5dip, 0, wmCRUDPaneFilter.Width - 10dip, filterFieldHeight)
	End If

	' Create the wmCRUDTableView1 columns
	colHdrList.Initialize
	rs = sqlMeta.ExecQuery2("SELECT `HeaderText` FROM `TableViewColumns` WHERE `TableName`=? ORDER BY `Position` ASC", Array As String(tablename))
	Do While rs.NextRow
		colHdrList.Add(rs.GetString("HeaderText"))
	Loop
	rs.Close
	wmCRUDTableView1.SetColumns(colHdrList)
	TableViewSetAllColumnsStyle(wmCRUDTableView1, "-fx-alignment: CENTER_LEFT;")

	colNum = -1
	filterNum = -1
	rs = sqlMeta.ExecQuery2("SELECT * FROM `TableViewColumns` WHERE `TableName`=? ORDER BY `Position` ASC", Array As String(tablename))
	Do While rs.NextRow
		colNum = colNum + 1
		If rs.GetString("SortOrder") <> "" Then
			dbSortOrder = rs.GetString("SortOrder")
			dbSortColumn = rs.GetString("QualifiedColumnName")
		End If
		wmCRUDTableView1.SetColumnsortable(colNum, False) ' We'll handle column sorting ourselves
		Try
			wmCRUDTableView1.SetColumnWidth(colNum, rs.GetDouble("Width"))
		Catch
			Log("") ' Just avoiding a compiler warning here
		End Try
		If myShowFilters Then
			Dim cbLst As List
			cbLst.Initialize
			cbLst = GetComboboxLists(rs.GetString("QualifiedColumnName"))
			filterNum = TV1SetColumns_CreateFilterField(filterNum, rs, cbLst.Get(0).As(List), cbLst.Get(1).As(List))
		End If
		' Set the 'Sort' arrow in the header cell
		Dim arrow As ImageView
		arrow.Initialize("")
		arrow.PickOnBounds = True
		Select Case rs.GetString("SortOrder")
			Case "COLLATE NOCASE ASC"
				arrow.SetImage(arrowUp)
			Case "COLLATE NOCASE DESC"
				arrow.SetImage(arrowDown)
			Case Else
				arrow.SetImage(arrowRight)
		End Select
		TableViewSetColumnGraphicOnly(wmCRUDTableView1, colNum, arrow)
	Loop
	rs.Close

	resizedColumnsList.Initialize
	resizedColumnsTimer.Enabled = False
	TV1SetColumnListeners ' To handle column resize events

	If myShowFilters And (showingDataAfterSettingFilter = False) Then
		If sqlMeta.ExecQuerySingleResult2("SELECT COUNT(*) FROM `TableViewColumns` WHERE (`TableName`=? And `Filtering`<>0)", Array As String(tablename)) > 0 Then
			SetDbWhereFilter(False)
			Sleep(10) ' Otherwise, events for the filter fields may unnecessarily fire after this method and an infinite loop occurs
		End If
	End If

	TV1SetColumnsInProgress = False

End Sub

Private Sub TV1SetColumns_CreateFilterField(filterNum As Int, rs As ResultSet, cbItemsDisplay As List, cbItemsValues As List) As Int

	Dim filterField As Node
	Dim filterFieldTag As String = rs.GetString("QualifiedColumnName")
	Dim filterFieldTextField As TextField
	Dim filterFieldCombobox As ComboBox
	Dim filterFieldDatePicker As DatePicker
	Dim filterFieldDatePicker2 As DatePicker
	Dim filterFieldIsComboBox As Boolean = IIf(rs.GetInt("DataType") = WMCRUD_DATATYPE_COMBOBOX, True, False)
	Dim filterFieldIsDatePicker As Boolean = IIf((rs.GetInt("DataType") > WMCRUD_DATATYPE_DATE_START) And (rs.GetInt("DataType") < WMCRUD_DATATYPE_DATE_END), True, False)
	Dim filterFieldBorderWidth As Int = borderWidth
	Dim filterDateFrom As String = ""
	Dim filterDateTo As String = ""
	Dim val As String = ""
	Dim filtering As Boolean = False
	Dim bordercolour As Paint = clr_foregroundPaint
	Dim toolTipSuffix As String = ""
	Dim i As Int

	filterNum = filterNum + 1

	If showingDataAfterSettingFilter Then
		If sqlMeta.ExecQuerySingleResult2("SELECT `Filtering` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, filterFieldTag)).As(Int) <> 0 Then
			filtering = True
		End If
		filterField = dbWhereFilterViewList.Get(filterNum)
		If filterField Is DatePicker Then
			filterFieldDatePicker = filterField
		Else If filterField Is ComboBox Then
			filterFieldCombobox = filterField
		Else
			filterFieldTextField = filterField
		End If
	Else If filterFieldIsDatePicker Then
		filterDateFrom = sqlMeta.ExecQuerySingleResult2("SELECT `FilterValueFrom` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, filterFieldTag)).As(String)
		filterDateTo = sqlMeta.ExecQuerySingleResult2("SELECT `FilterValueTo` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, filterFieldTag)).As(String)
		val = filterDateFrom & filterDateTo
		' From
		filterFieldDatePicker.Initialize("FilterFieldDP")
		filterFieldDatePicker.Editable = False
		filterFieldDatePicker.Id = "from"
		If filterDateFrom = "" Then
			filterFieldDatePicker.DateTicks = 0
		Else
			filterFieldDatePicker.DateTicks = YYYYMMDDtoDate(filterDateFrom)
		End If
		dbWhereFilterViewList.Add(filterFieldDatePicker)
		filterField = filterFieldDatePicker
		toolTipSuffix = Localise("From")
		' To
		filterFieldDatePicker2.Initialize("FilterFieldDP")
		filterFieldDatePicker2.Editable = False
		filterFieldDatePicker2.Id = "to"
		If filterDateFrom = "" Then
			filterFieldDatePicker2.DateTicks = 0
		Else
			filterFieldDatePicker2.DateTicks = YYYYMMDDtoDate(filterDateFrom)
		End If
		dbWhereFilterViewList.Add(filterFieldDatePicker2)
	Else If filterFieldIsComboBox Then
		filterFieldCombobox.Initialize("FilterFieldCB")
		filterFieldCombobox.As(B4XView).Font = arialGeneral.As(B4XFont)
		filterFieldCombobox.Items.AddAll(cbItemsDisplay)
		val = rs.GetString("FilterValueFrom")
		If val <> "" Then
			i = cbItemsValues.IndexOf(val)
			If i >= 0 Then filterFieldCombobox.SelectedIndex = i
		End If
		dbWhereFilterViewList.Add(filterFieldCombobox)
		filterField = filterFieldCombobox
	Else
		filterFieldTextField.Initialize("FilterFieldTF")
		filterFieldTextField.As(B4XView).Font = arialGeneral.As(B4XFont)
		filterFieldTextField.PromptText = rs.GetString("HeaderText")
		val = rs.GetString("FilterValueFrom")
		If val <> "" Then filterFieldTextField.Text = val
		dbWhereFilterViewList.Add(filterFieldTextField)
		filterField = filterFieldTextField
		toolTipSuffix = Localise("Enter (part of) a value and press ENTER to apply this filter") & ". " & _
						Localise("Use two double quotes ("") to filter on empty text values, or zero for empty numeric values" & ".")
	End If

	TV1SetColumns_CreateFilterField_AddNode(filterField, filterNum, rs, toolTipSuffix)
	If filterFieldDatePicker2.IsInitialized Then
		filterNum = filterNum + 1
		TV1SetColumns_CreateFilterField_AddNode(filterFieldDatePicker2, filterNum, rs, Localise("To"))
	End If

	If filtering Or (val <> "") Then
		lblFilters.As(B4XView).TextColor = xui.Color_Red
		bordercolour = fx.Colors.Red
		filterFieldBorderWidth = 2dip
	End If

	CSSUtils.SetBorder(filterField, filterFieldBorderWidth, bordercolour, 0)

	Return filterNum

End Sub

Private Sub TV1SetColumns_CreateFilterField_AddNode(filterField As Node, filterNum As Int, rs As ResultSet, toolTipSuffix As String)

	wmCRUDPaneFilter.AddNode(filterField, 5dip, (filterNum + 1) * (filterFieldHeight + 1dip), wmCRUDPaneFilter.Width - 10dip, filterFieldHeight)
	filterField.PickOnBounds = True
	filterField.Tag = rs.GetString("QualifiedColumnName")
	SetToolTip(filterField, rs.GetString("HeaderText") & IIf(toolTipSuffix = "", "", " (" & toolTipSuffix & ")"), arialBoldGeneral, True)

End Sub

Private Sub wmCRUDTableView1MouseClicked_Filter(e As Event)

	' Code based on https://www.b4x.com/android/forum/threads/sorting-tableview-custom-columns.65292/

	Dim jo As JavaObject = e
	Dim v As Node = jo.RunMethod("getTarget", Null)

	If showingData Then Return

	Do While v.IsInitialized
		If GetType(v).EndsWith(".TableColumnHeader") Then
			jo = v
			Dim TableColumn As JavaObject = jo.RunMethod("getTableColumn", Null)
			Dim colHdrText As String = TableColumn.RunMethod("getText", Null)
			If TableViewGetColNumFromHeaderText(wmCRUDTableView1, colHdrText) >= 0 Then
				Dim colSortOrder As String = sqlMeta.ExecQuerySingleResult2("SELECT `SortOrder` FROM `TableViewColumns` WHERE `TableName`=? AND `HeaderText`=?", Array As String(tablename, colHdrText))
				Dim newSortOrder As String = ""
				Select Case colSortOrder
					Case "COLLATE NOCASE DESC"
						newSortOrder = ""
					Case ""
						newSortOrder = "COLLATE NOCASE ASC"
					Case "COLLATE NOCASE ASC"
						newSortOrder = "COLLATE NOCASE DESC"
				End Select
				sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `SortOrder`=? WHERE `TableName`=?", Array As String("", tablename)) ' Reset existing sort order
				sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `SortOrder`=? WHERE `TableName`=? AND `HeaderText`=?", Array As String(newSortOrder, tablename, colHdrText))
				ShowData(Null, False)
			Else
				Log("wmCRUD.wmCRUDTableView1MouseClicked_Filter (" & tablename & ") - column not found for table '" & tablename & "', header: *" & colHdrText & "*")
			End If
			e.Consume
			Return
		End If
		v = v.Parent
	Loop

End Sub

Private Sub TvLbl_MouseClicked(EventData As MouseEvent)

	' Code based on https://www.b4x.com/android/forum/threads/get-contextmenus-working-on-nodes-controls-which-do-not-support-it.102721

	If EventData.SecondaryButtonPressed = False Then Return ' Not a right-click

	Dim lstItems As List
	Dim lbl As B4XView = Sender
	Dim lblTag As wmCRUDlabelTagType = lbl.Tag

	lstItems.Initialize
	contextMenu1.MenuItems.Clear

	If allowNew Then
		lstItems.Add(Localise("Duplicate"))
		lstItems.Add(Localise("New"))
	End If

	If TableViewGetSelectedRows(wmCRUDTableView1).Size > 1 Then
		If allowDelete Then lstItems.Add(Localise("Delete"))
	Else
		If allowEdit And (cannotEditIDvalues.IndexOf(lblTag.ID) < 0) Then lstItems.Add(Localise("Edit"))
		If allowDelete And (cannotDeleteIDvalues.IndexOf(lblTag.ID) < 0) Then lstItems.Add(Localise("Delete"))
	End If

	lstItems.SortCaseInsensitive(True)
	For Each oneItem As String In lstItems
		ContextMenu1_AddItem(oneItem)
	Next

	If contextMenu1.MenuItems.Size > 0 Then
		contextMenu1.MenuItems.Add(GetMenuSeparatorItem)
		ContextMenu1_AddItem(Localise("Cancel"))
	End If

	' Trap the mouseclicked event, get the necessary data and show the ContextMenu
	Private joCM As JavaObject = contextMenu1
	Private joED As JavaObject = EventData
	Private ScreenX As Double = joED.RunMethod("getScreenX", Null)
	Private ScreenY As Double = joED.RunMethod("getScreenY", Null)
	joCM.RunMethod("show", Array(lbl, ScreenX, ScreenY))

End Sub

Private Sub ContextMenu1_AddItem(txt As String)

	' Code based on https://www.b4x.com/android/forum/threads/get-contextmenus-working-on-nodes-controls-which-do-not-support-it.102721

	Private TheMenuItem As MenuItem
	TheMenuItem.Initialize(txt, "ContextMenu1")
	contextMenu1.MenuItems.Add(TheMenuItem)

End Sub

Private Sub ContextMenu1_Action

	' Code based on https://www.b4x.com/android/forum/threads/get-contextmenus-working-on-nodes-controls-which-do-not-support-it.102721

	Private menuItem1 As MenuItem = Sender

	Select Case menuItem1.Text
		Case Localise("Cancel")
			Return
		Case Localise("Delete")
			DeleteRows
		Case Localise("Duplicate")
			DuplicateRows
		Case Localise("Edit")
			Update_Or_Add(False)
		Case Localise("New")
			Update_Or_Add(True)
	End Select

End Sub
#End Region ' TableView

#Region Filters
Private Sub LblFilters_MouseClicked(EventData As MouseEvent)

	ResetDbWhereFilters(True)
	ShowData(Null, False)

End Sub

Private Sub ResetDbWhereFilters(clearDbValues As Boolean)

	dbWhereFilter = ""
	dbWhereFilterValuesList.Initialize
	dbWhereFilterViewList.Initialize

	If clearDbValues Then
		sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Filtering`=0, `FilterValueFrom`='', `FilterValueTo`='' WHERE `TableName`=?;", _
								Array As String(tablename))
	End If

End Sub

Private Sub FilterFieldDP_ValueChanged(Value As Long)

	Dim dp As DatePicker = Sender
	Dim dpTag As String = dp.Tag.As(String)
	Dim colName As String = IIf(dp.Id = "from", "FilterValueFrom", "FilterValueTo")
	Dim otherColName As String = IIf(dp.Id = "to", "FilterValueFrom", "FilterValueTo")
	Dim newDate As String = ""
	Dim filtering As Int = 0
	Dim s As String

	s = sqlMeta.ExecQuerySingleResult2("SELECT `" & colName & "` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, dpTag)).As(String)
	If s = "" Then
		dp.DateTicks = 0
	Else
		dp.DateTicks = YYYYMMDDtoDate(s)
	End If

	If dp.DateTicks <> 0 Then newDate = DateToYYYYMMDD(dp.DateTicks, "")

	s = sqlMeta.ExecQuerySingleResult2("SELECT `" & otherColName & "` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, dpTag)).As(String)
	filtering = IIf((newDate = "") And (s = ""), 0, 1)

	sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Filtering`=?, `" & colName & "`=?, WHERE `TableName`=? AND `QualifiedColumnName`=?;", _
							Array As String(filtering, newDate, tablename, dpTag))

	SetDbWhereFilter(True)

End Sub

Private Sub FilterFieldTF_Action

	If TV1SetColumnsInProgress = False Then SetDbWhereFilter(True)

End Sub

Private Sub FilterFieldCB_ValueChanged(Value As Object)

	If TV1SetColumnsInProgress = False Then SetDbWhereFilter(True)

End Sub

Private Sub SetDbWhereFilter(filterFieldClicked As Boolean)

	Dim i As Int
	Dim s As String

	dbWhereFilter = ""
	dbWhereFilterValuesList.Initialize

	For i = 0 To (dbWhereFilterViewList.Size - 1)
		If dbWhereFilterViewList.Get(i) Is TextField Then
			Dim tf As TextField = dbWhereFilterViewList.Get(i)
			If tf.Text = QUOTE & QUOTE Then
				If dbWhereFilter <> "" Then dbWhereFilter = dbWhereFilter & " AND "
				dbWhereFilter = dbWhereFilter & tf.Tag.As(String) & "=?"
				dbWhereFilterValuesList.Add("")
			Else If tf.Text <> "" Then
				If dbWhereFilter <> "" Then dbWhereFilter = dbWhereFilter & " AND "
				dbWhereFilter = dbWhereFilter & tf.Tag.As(String) & " LIKE ?"
				dbWhereFilterValuesList.Add("%" & tf.Text & "%")
			End If
			If filterFieldClicked Then
				sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Filtering`=?, `FilterValueFrom`=? WHERE `TableName`=? AND `QualifiedColumnName`=?;", _
										Array As String(IIf(tf.Text = "", 0, 1), tf.Text, tablename, tf.Tag.As(String)))
			End If
		Else If dbWhereFilterViewList.Get(i) Is ComboBox Then
			Dim cb As ComboBox = dbWhereFilterViewList.Get(i)
			Dim cbValue As String = GetComboboxDbValue(cb.Tag.As(String), cb.Value.As(String))
			If cbValue <> "" Then
				If dbWhereFilter <> "" Then dbWhereFilter = dbWhereFilter & " AND "
				dbWhereFilter = dbWhereFilter & cb.Tag.As(String) & " = ?"
				dbWhereFilterValuesList.Add(cbValue)
			End If
			If filterFieldClicked Then
				sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Filtering`=?, `FilterValueFrom`=? WHERE `TableName`=? AND `QualifiedColumnName`=?;", _
										Array As String(IIf(cbValue = "", 0, 1), cbValue, tablename, cb.Tag.As(String)))
			End If
		Else
			' Note: values were saved to the meta database in Sub FilterFieldDP_ValueChanged
			Dim dp As DatePicker = dbWhereFilterViewList.Get(i)
			Dim dpTag As String = dp.Tag.As(String)
			Dim dpFilterValueColName As String = IIf(dp.Id = "from", "FilterValueFrom", "FilterValueTo")
			Dim dpFilterValueOperator As String = IIf(dp.Id = "from", " >= ?", " <= ?")
			s = sqlMeta.ExecQuerySingleResult2("SELECT `" & dpFilterValueColName & "` FROM `TableViewColumns` WHERE `TableName`=? AND `QualifiedColumnName`=?", Array As String(tablename, dpTag)).As(String)
			If s <> "" Then
				If dbWhereFilter <> "" Then dbWhereFilter = dbWhereFilter & " AND "
				dbWhereFilter = dbWhereFilter & dpTag & dpFilterValueOperator
				dbWhereFilterValuesList.Add(s)
			End If
		End If
	Next

	If dbWhereFilter <> "" Then
		dbWhereFilter = " WHERE (" & dbWhereFilter & ")"
		If filterFieldClicked Then showingDataAfterSettingFilter = True
	Else
		If filterFieldClicked Then showingDataAfterSettingFilter = False
	End If

	If filterFieldClicked Then ShowData(Null, False)

	Sleep(10) ' Otherwise, events for the filter fields may unnecessarily fire after this method and an infinite loop occurs

End Sub
#End Region ' Filters

#Region TableView helpers
Private Sub GetComboboxLists(qualifiedColName As String) As List

	Dim cbLst As List
	Dim dispLst As List
	Dim valLst As List

	cbLst.Initialize
	dispLst.Initialize
	valLst.Initialize

	If mapColDefsCBlist.ContainsKey(qualifiedColName.ToLowerCase) Then
		cbLst = mapColDefsCBlist.Get(qualifiedColName.ToLowerCase)
	Else
		cbLst.Add(dispLst)
		cbLst.Add(valLst)
	End If

	Return cbLst

End Sub

Private Sub GetComboboxDisplayValue(qualifiedColName As String, dbValue As String) As String

	Dim cbLst As List = GetComboboxLists(qualifiedColName)
	Dim cbLstDisplay As List = cbLst.Get(0).As(List)
	Dim cbLstValues As List = cbLst.Get(1).As(List)
	Dim i As Int = cbLstValues.IndexOf(dbValue)
	Return IIf(i < 0, "", cbLstDisplay.Get(i).As(String))

End Sub

Private Sub GetComboboxDbValue(qualifiedColName As String, displayValue As String) As String

	Dim cbLst As List = GetComboboxLists(qualifiedColName)
	Dim cbLstDisplay As List = cbLst.Get(0).As(List)
	Dim cbLstValues As List = cbLst.Get(1).As(List)
	Dim i As Int = cbLstDisplay.IndexOf(displayValue)
	Return IIf(i < 0, "", cbLstValues.Get(i).As(String))

End Sub

Private Sub TV1GetLabelFromRow(cell As Object) As Label

	Dim v As B4XView = cell

	If v Is Pane Then
		Return v.GetView(0)
	Else
		Return v
	End If

End Sub

Private Sub TV1GetColWidthFromColNum(colNum As Int) As Double

	Return sqlMeta.ExecQuerySingleResult2("SELECT `Width` FROM `TableViewColumns` WHERE `TableName`=? AND `Position`=?", Array As String(tablename, colNum)).As(Double)

End Sub

Private Sub TV1SetFAlabel(colNum As Int, lbl1 As Label, FAsymbol As String, txtColour As Paint, tooltip As String) As Label

	' Based on https://www.b4x.com/android/forum/threads/fontawesome-at-runtime.85628/

	Dim lbl As Label

	If lbl1.IsInitialized Then
		lbl = lbl1
	Else
		lbl.Initialize("TvLbl")
		lbl.As(B4XView).Width = wmCRUDTableView1.GetColumnWidth(colNum)
		lbl.As(B4XView).Height = tvCellSize
	End If

	RemoveTooltip(lbl)
	lbl.Font = fontAwesomeForTableView
	lbl.Text = FAsymbol
	lbl.As(B4XView).Color = clr_backgroundInt
	lbl.TextColor = txtColour
	lbl.TooltipText = tooltip
	lbl.PickOnBounds = True
	lbl.Alignment = "CENTER"
	SetToolTipFont(lbl, arialBoldGeneral)

	Return lbl

End Sub

Private Sub TV1SetMIlabel(colNum As Int, lbl1 As Label, MIsymbol As String, txtColour As Paint, tooltip As String) As Label

	' Based on https://www.b4x.com/android/forum/threads/fontawesome-at-runtime.85628/

	Dim lbl As Label

	If lbl1.IsInitialized Then
		lbl = lbl1
	Else
		lbl.Initialize("TvLbl")
		lbl.As(B4XView).Width = wmCRUDTableView1.GetColumnWidth(colNum)
		lbl.As(B4XView).Height = tvCellSize
	End If

	RemoveTooltip(lbl)
	lbl.Font = materialIconsForTableView
	lbl.Text = MIsymbol
	lbl.As(B4XView).Color = clr_backgroundInt
	lbl.TextColor = txtColour
	lbl.TooltipText = tooltip
	lbl.PickOnBounds = True
	lbl.Alignment = "CENTER"
	SetToolTipFont(lbl, arialBoldGeneral)

	Return lbl

End Sub

Private Sub wmCRUDTableView1_Resize(Width As Double, Height As Double)

	If resizingColumns Or TV1SetColumnsInProgress Or (tablename = "") Then Return

	Dim w As Double = (Width - 30dip) / wmCRUDTableView1.ColumnsCount

	resizingColumns = True

	For i = 0 To wmCRUDTableView1.ColumnsCount - 1
		Try
			wmCRUDTableView1.SetColumnWidth(i, TV1GetColWidthFromColNum(i))
		Catch
			wmCRUDTableView1.SetColumnWidth(i, w)
		End Try
	Next

	resizingColumns = False

End Sub

Private Sub TV1GetSelectedIDs(listSelectedRowsOrNull As List, sortAsc As Boolean) As List

	Dim listSelectedRows As List
	Dim listSelectedIDs As List

	If listSelectedRowsOrNull = Null Then
		listSelectedRows = TableViewGetSelectedRows(wmCRUDTableView1)
	Else
		listSelectedRows = listSelectedRowsOrNull
	End If

	listSelectedIDs.Initialize

	For i = 0 To (listSelectedRows.Size - 1)
		Dim row() As Object = listSelectedRows.Get(i)
		Dim lbl As Label = TV1GetLabelFromRow(row(0))
		Dim lblTag As wmCRUDlabelTagType = lbl.Tag
		listSelectedIDs.Add(lblTag.ID)
	Next

	listSelectedIDs.Sort(sortAsc)
	Return listSelectedIDs

End Sub

Private Sub TV1SaveColumnWidthsAndOrder(savePositions As Boolean)

	Dim i As Int

	If (wmCRUDPaneContent.Visible = False) Or (wmCRUDTableView1.Visible = False) Then Return

	' Positions should NOT be saved when wmCRUDTableView1 has already been setup,
	' as reordering columns by dragging them around will NOT change the order of
	' the data in rows that are fetched from the TableView!
	For i = 0 To (wmCRUDTableView1.ColumnsCount - 1)
		If savePositions Then
			sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Position`=?, `Width`=? WHERE `TableName`=? AND `HeaderText`=?", Array As String(i, wmCRUDTableView1.GetColumnWidth(i), tablename, wmCRUDTableView1.GetColumnHeader(i)))
		Else
			sqlMeta.ExecNonQuery2("UPDATE `TableViewColumns` SET `Width`=? WHERE `TableName`=? AND `HeaderText`=?", Array As String(wmCRUDTableView1.GetColumnWidth(i), tablename, wmCRUDTableView1.GetColumnHeader(i)))
		End If
	Next

End Sub

Private Sub TV1SetColumnListeners

	' Code from https://www.b4x.com/android/forum/threads/solved-tableview-column_resized-event.127230/#post-796527

	Dim jo As JavaObject = wmCRUDTableView1
	Dim ColumnList As List = jo.RunMethodJO("getColumns", Null)
    
	For Each Col As JavaObject In ColumnList
		Dim Prop As JavaObject = Col.RunMethod("widthProperty",Null)
		Dim o As Object = Prop.CreateEventFromUI("javafx.beans.value.ChangeListener", "ColumnChanged",Null)
		Prop.RunMethod("addListener",Array(o))
	Next

End Sub

Private Sub ColumnChanged_Event(MethodName As String, Args() As Object) As Object

	' Code based on https://www.b4x.com/android/forum/threads/solved-tableview-column_resized-event.127230/#post-796527

	If resizingColumns Or TV1SetColumnsInProgress Then Return Null

	Dim col As JavaObject = Sender
	Dim jo As JavaObject = wmCRUDTableView1
	Dim columnList As List = jo.RunMethodJO("getColumns", Null)
	Dim colNum As Int = columnList.IndexOf(col.RunMethod("getBean",Null))
	'Dim Width As Double = Args(2)
	'Log(ColNum & " : " & Width)

	If resizedColumnsList.IndexOf(colNum) < 0 Then resizedColumnsList.Add(colNum)

	resizedColumnsTimer.Enabled = False
	resizedColumnsTimer.Interval = 500 ' If, after half a second, resizing activity stops, apply it to the labels in the resized columns in wmCRUDTableView1's rows
	resizedColumnsTimer.Enabled = True

	Return Null

End Sub

Private Sub ResizedColumnsTimer_Tick

	' If, after half a second, resizing activity stops, apply it to the labels in the resized columns in wmCRUDTableView1's rows

	Dim i As Int
	Dim i2 As Int

	resizedColumnsTimer.Enabled = False
	resizingColumns = True

	For i = 0 To wmCRUDTableView1.Items.Size - 1
		Dim row() As Object = wmCRUDTableView1.Items.Get(i)
		For i2 = 0 To (resizedColumnsList.Size - 1)
			Dim iCol As Int = resizedColumnsList.Get(i2)
			Dim lbl As Label = TV1GetLabelFromRow(row(iCol))
			lbl.As(B4XView).Width = wmCRUDTableView1.GetColumnWidth(iCol)
		Next
	Next

	resizedColumnsList.Clear
	resizingColumns = False

	Sleep(0)

	TV1SaveColumnWidthsAndOrder(False)

End Sub

Private Sub wmCRUDTableView1MouseAny_Filter(e As Event)

	' Code based on https://www.b4x.com/android/forum/threads/tableview-resize-column-event.67650/#post-428670
	' Disables double-clicking the TableView's headers to resize the columns, as a bug in JavaFX then hides that column's content

	Dim jo As JavaObject = e
	Dim v As Node = jo.RunMethod("getTarget", Null)

	If showingData Then Return

	If GetType(v) = "javafx.scene.shape.Rectangle" Then
		Dim m As MouseEvent = e
		If m.clickCount > 1 Then m.Consume
	End If

End Sub

Private Sub SetColumnDefinitionsOrderedLikeTV

	Dim rs As ResultSet

	columnDefinitionsOrderedLikeTV.Initialize

	rs = sqlMeta.ExecQuery2("SELECT * FROM `TableViewColumns` WHERE `TableName`=? ORDER BY `Position` ASC", Array As String(tablename))
	Do While rs.NextRow
		For Each oneColDef As wmCRUDcolumnDefinition In columnDefinitions
			If oneColDef.qualifiedColumnName = rs.GetString("QualifiedColumnName") Then
				columnDefinitionsOrderedLikeTV.Add(oneColDef)
			End If
		Next
	Loop
	rs.Close

End Sub
#End Region ' TableView helpers

#Region Utility
Private Sub SetWaitCursor(waitCursor As Boolean)

	Try
		wmCRUDLabelWorking.Text = Localise("Working...") ' In case the text was changed after Initialize2
		BlurPane(wmCRUDPaneContent, waitCursor)
		If waitCursor Then
			wmCRUDLabelWorking.Left = (MainForm.Width - wmCRUDLabelWorking.Width) / 2
			wmCRUDLabelWorking.top = (MainForm.Height - wmCRUDLabelWorking.Height) / 2
		End If
		wmCRUDLabelWorking.Visible = waitCursor
		BringToFront(wmCRUDLabelWorking)
		If waitCursor Then
			MainForm.RootPane.MouseCursor = fx.Cursors.WAIT
		Else
			MainForm.RootPane.MouseCursor = fx.Cursors.DEFAULT
		End If
	Catch
		Log("wmCRUD.SetWaitCursor (" & tablename & "): " & LastException)
	End Try

End Sub

' Library dependencies: JavaObject
Private Sub BlurPane(p As Pane, blurIt As Boolean)

	' Code from https://www.b4x.com/android/forum/threads/blur-mainform-on-opening-another-form.86778/

	Dim Blur As JavaObject
	Blur.InitializeNewInstance("javafx.scene.effect.BoxBlur",Null)

	If blurIt Then
		Blur.RunMethod("setWidth",Array(4.0D))
		Blur.RunMethod("setHeight",Array(4.0D))
		Blur.RunMethod("setIterations",Array(2))
	Else
		Blur.RunMethod("setWidth",Array(0.0D))
		Blur.RunMethod("setHeight",Array(0.0D))
		Blur.RunMethod("setIterations",Array(0))
	End If

	p.As(JavaObject).RunMethod("setEffect",Array(Blur))

End Sub

' Library dependencies: JavaObject
Private Sub BringToFront(n As Node)

	' Code from https://www.b4x.com/android/forum/threads/pane-bringtofront.70149

	Dim joNode As JavaObject = n
	joNode.RunMethod("toFront", Null)

End Sub

Private Sub IsSignedInteger(s As String, allowNullString As Boolean) As Boolean

	' Regex expression based on http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Dim sNoThousands As String = s.Replace(thousandsSeparator, "")
		Return IsNumber(sNoThousands) And Regex.IsMatch("^(\+|-)?\d+$", sNoThousands)
	End If

End Sub

Private Sub IsSignedDecimal(s As String, allowNullString As Boolean, decSep As String) As Boolean

	Dim sep As String
	If decSep.Trim = "" Then
		sep = decimalSeparator
	Else
		sep = decSep
	End If

	' Regex expression based on https://community.oracle.com/thread/429716
	' combined with http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Dim sNoThousands As String = s.Replace(thousandsSeparator, "")
		Return IsNumber(sNoThousands) And Regex.IsMatch("^(\+|-)?(\" & sep & "[0-9]+|[0-9]+(\" & sep & "[0-9]*)*)$", sNoThousands)
	End If

End Sub

' Library dependencies: JavaObject
Private Sub SetToolTipFont(N As Control, F As Font) As Boolean

	' Based on https://www.b4x.com/android/forum/threads/consistent-tooltips.100450/

	Try
		Dim JoN As JavaObject = N
		JoN.RunMethodJO("getTooltip",Null).RunMethod("setFont",Array (F))
		Return True
	Catch
		Return False
	End Try

End Sub

Private Sub RemoveTooltip(n As Node)

	SetToolTip(n, "", Null, False)

End Sub

' Library dependencies: JavaObject
Private Sub SetToolTip(Nd As Node, msg As String, fnt As Font, add As Boolean)

	' Code based on https://www.b4x.com/android/forum/threads/add-a-tooltip-to-an-imageview.65439/

	Dim joToolTip As JavaObject
	Dim joToolTip2 As JavaObject = joToolTip.InitializeNewInstance("javafx.scene.control.Tooltip", Array(msg))
	If add = True Then
		joToolTip.RunMethod("install", Array(Nd, joToolTip2))
		If fnt.IsInitialized And (fnt <> Null) Then joToolTip2.RunMethod("setFont", Array(fnt))
	Else
		joToolTip.RunMethod("uninstall", Array(Nd, joToolTip2))
	End If

End Sub

Private Sub NullOrEmptyStringToDefault(s As String, deflt As Object) As String

	If s = Null Then
		Return deflt
	Else If s = "" Then
		Return deflt
	Else If s.ToLowerCase = "null" Then
		Return deflt
	Else
		Return s
	End If

End Sub

' Library dependencies: Reflection
Private Sub FileExistsInDirAssets(fname As String) As Boolean

	' Code based on https://www.b4x.com/android/forum/threads/file-exists-on-dirassets.138614

	Dim r As Reflector
	r.Target = Me
	r.Target = r.RunMethod("getClass")
	Dim In As InputStream = r.RunMethod2("getResourceAsStream", "/Files/" &  fname, "java.lang.String")
	If In.IsInitialized = False Then Return False
	In.Close
	Return True

End Sub

' Library dependencies: JavaObject
Private Sub TableViewGetSelectedRows(TV As TableView) As List

	'Code based on https://www.b4x.com/android/forum/threads/tableview-select-multiple-rows.60161

	Dim jo As JavaObject = TV
	Dim sm As JavaObject
	sm = jo.RunMethod("getSelectionModel",Null)
	Dim listSelectedRows As List
	listSelectedRows.Initialize
	listSelectedRows.AddAll(sm.RunMethodJO("getSelectedItems", Null))
	Return listSelectedRows

End Sub

Private Sub DateToYYYYMMDD(dateIn As Long, sep As String) As String

	Dim OrigFormat As String = DateTime.DateFormat 'save orig date format
	DateTime.DateFormat = "yyyy" & sep & "MM" & sep & "dd"
	Dim YYYYMMDDout As String = DateTime.Date(dateIn)
	DateTime.DateFormat = OrigFormat 'return to orig date format
	Return YYYYMMDDout

End Sub

Private Sub DDMMYYYYtoDate(DDMMYYYYin As String, sep As String) As Long

	Dim OrigFormat As String = DateTime.DateFormat 'save orig date format
	DateTime.DateFormat = "dd" & sep & "MM" & sep & "yyyy"
	Dim theDate As String = IIf(DDMMYYYYin.StartsWith("'"), DDMMYYYYin.SubString(1), DDMMYYYYin) ' For Excel values that were stored as text
	Dim dateOut As Long = DateTime.DateParse(theDate)
	DateTime.DateFormat = OrigFormat 'return to orig date format
	Return dateOut

End Sub

Private Sub YYYYMMDDtoDate(YYYYMMDDin As String) As Long

	Dim OrigFormat As String = DateTime.DateFormat 'save orig date format
	DateTime.DateFormat = "yyyyMMdd"
	Dim dateOut As Long = DateTime.DateParse(YYYYMMDDin)
	DateTime.DateFormat = OrigFormat 'return to orig date format
	Return dateOut

End Sub

' Library dependencies: (j)SQL
Private Sub SQLtableExists(theSQLdb As SQL, tablename1 As String) As Boolean

	' SQLite - code based on DBUTILS 2 : https://www.b4x.com/android/forum/threads/b4x-dbutils-2.81280/
	Try
		Dim count As Int = theSQLdb.ExecQuerySingleResult2("SELECT count(name) FROM sqlite_master WHERE type='table' AND name=? COLLATE NOCASE", Array As String(tablename1))
		Return count > 0
	Catch
		Log("") ' Just avoiding a compiler warning here
	End Try

	' Non-SQLite
	' THIS HAS ONLY BEEN TESTED WITH MYSQL; the database must adhere to ANSI standards in order for this to work.
	' Code from https://stackoverflow.com/questions/464474/check-if-a-sql-table-exists
	Try
		Dim count As Int = theSQLdb.ExecQuerySingleResult2("Select count(*) from information_schema.tables where table_name=?", Array As String(tablename1))
		Return count > 0
	Catch
		Log("wmSQLtableSync.SQLtableExists: " & LastException)
		Return False
	End Try

End Sub

' Library dependencies: (j)SQL
Private Sub SQLcolumnExists(theSQLdb As SQL, tablename1 As String, columnName As String) As Boolean

	If SQLtableExists(theSQLdb, tablename) = False Then Return False

	Dim rs As ResultSet = theSQLdb.ExecQuery("SELECT * FROM `" & tablename1 & "` LIMIT 1")
	Dim i As Int
	Dim exists As Boolean = False

	For i = 0 To (rs.ColumnCount - 1)
		If columnName.EqualsIgnoreCase(rs.GetColumnName(i)) Then
			exists = True
			Exit
		End If
	Next

	rs.Close
	Return exists

End Sub

' Library dependencies: (j)SQL
Private Sub SQLgetTableColumnNames(theSQLdb As SQL, tblName As String, sortAsc As Boolean) As List

	Dim lst As List
	lst.Initialize

	' SQLite
	Try
		Dim rs As ResultSet = theSQLdb.ExecQuery("PRAGMA table_info(" & tblName & ")")
		Do While rs.NextRow
			lst.Add(rs.GetString("name"))
		Loop
		rs.Close
		If sortAsc Then lst.SortCaseInsensitive(True)
		Return lst
	Catch
		Log("") ' Just avoiding a compiler warning here
	End Try

	' Non-SQLite
	' THIS HAS ONLY BEEN TESTED WITH MYSQL; the database must adhere to ANSI standards in order for this to work.
	Try
		Dim rs As ResultSet = theSQLdb.ExecQuery2("Select column_name from information_schema.columns where table_name=?", Array As String(tblName))
		Do While rs.NextRow
			lst.Add(rs.GetString("column_name"))
		Loop
		rs.Close
		If sortAsc Then lst.SortCaseInsensitive(True)
		Return lst
	Catch
		Log("wmSQLtableSync.SQLgetTableColumnNames: " & LastException)
		If rs.IsInitialized Then rs.Close
		Return lst
	End Try

End Sub

Private Sub TableViewGetColNumFromHeaderText(TV As TableView, HeaderText As String) As Int

	Dim i As Int

	For i = 0 To (TV.ColumnsCount - 1)
		If TV.GetColumnHeader(i) = HeaderText Then Return i
	Next

	Return -1

End Sub

' Set the image for the arrow in a TableView column header
' xNode is an ImageView
' Library dependencies: JavaObject
Private Sub TableViewSetColumnGraphicOnly(TV As TableView, colIndex As Int, xNode As Node)

	' Code from https://www.b4x.com/android/forum/threads/tableview-adding-tooltips-to-column-headers.96336

	Dim jo As JavaObject = TV
	Dim column As JavaObject = jo.RunMethodJO("getColumns", Null).RunMethod("get", Array(colIndex))
	column.RunMethod("setGraphic", Array(xNode))

End Sub

' Style is e.g. "-fx-alignment: CENTER_LEFT;"
' Library dependencies: none
Private Sub TableViewSetAllColumnsStyle(TV As TableView, Style As String)

	For i = 0 To (TV.ColumnsCount - 1)
		TableViewSetColumnStyle(TV, i, Style)
	Next

End Sub

' Style is e.g. "-fx-alignment: CENTER_LEFT;"
' Library dependencies: JavaObject
Private Sub TableViewSetColumnStyle(TV As TableView, Index As Int, Style As String)

	Dim jo As JavaObject = TV
	Dim Column As JavaObject = jo.RunMethodJO("getColumns", Null).RunMethod("get", Array(Index))
	Column.RunMethod("setStyle", Array(Style))

End Sub

' Change (or hide) TableView "No content in table" text
' Library dependencies: JavaObject
Private Sub SetTableViewPlaceHolderText(TV As TableView, Text As String)

	' Code from https://www.b4x.com/android/forum/threads/tableview-replace-the-default-no-content-in-table-for-empty-table.82744/

	Dim TVJO As JavaObject = TV
	Dim L As Label
	L.Initialize("")
	L.Text = Text
	TVJO.RunMethod("setPlaceholder",Array(L))

End Sub

Private Sub ColorToHexCSS(clr As Int) As String

	' Code based on https://www.b4x.com/android/forum/threads/b4x-hextocolor-and-colortohex.114300

	Dim bc As ByteConverter
	Return "#" & bc.HexFromBytes(bc.IntsToBytes(Array As Int(clr))).SubString(2)

End Sub

Private Sub CopyList(source As List, asString As Boolean) As List

	Dim lst As List
	lst.Initialize
	If source.IsInitialized Then
		For Each o As Object In source
			lst.Add(IIf(asString, o.As(String), o))
		Next
	End If
	Return lst

End Sub

' Use as: [context]menu1.MenuItems.Add(GetMenuSeparatorItem)
' Library dependencies: JavaObject
Private Sub GetMenuSeparatorItem As Object

	' Code from https://www.b4x.com/android/forum/threads/how-to-make-a-menubar.114752

	Dim TJO As JavaObject
	TJO.InitializeNewInstance("javafx.scene.control.SeparatorMenuItem", Null)
	Return TJO

End Sub
#End Region ' Utility